#    Copyright 2017 FUJITSU LIMITED
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from osc_lib.cli import identity as identity_utils
from osc_lib import exceptions
from osc_lib import utils
from osc_lib.utils import columns as column_util

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


_attr_map = [
    ('id', 'ID', column_util.LIST_BOTH),
    ('name', 'Name', column_util.LIST_BOTH),
    ('type', 'Type', column_util.LIST_BOTH),
    ('endpoints', 'Endpoints', column_util.LIST_BOTH),
    ('description', 'Description', column_util.LIST_LONG_ONLY),
    ('project_id', 'Project', column_util.LIST_LONG_ONLY),
]

_attr_map_dict = {
    'id': 'ID',
    'name': 'Name',
    'type': 'Type',
    'endpoints': 'Endpoints',
    'description': 'Description',
    'project_id': 'Project',
}


class CreateEndpointGroup(command.ShowOne):
    _description = _("Create an endpoint group")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description for the endpoint group'),
        )
        parser.add_argument(
            'name', metavar='<name>', help=_('Name for the endpoint group')
        )
        parser.add_argument(
            '--type',
            required=True,
            choices=['subnet', 'cidr'],
            help=_(
                'Type of endpoints in group (e.g. subnet, cidr, network, '
                'router). Currently only subnet and cidr are supported.'
            ),
        )
        parser.add_argument(
            '--value',
            action='append',
            dest='endpoints',
            required=True,
            help=_(
                'Endpoint(s) for the group. Must all be of the same type. '
                '(--value) option can be repeated'
            ),
        )
        identity_utils.add_project_owner_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        attrs: dict[str, Any] = {}
        if parsed_args.project is not None:
            identity_client = self.app.client_manager.sdk_connection.identity
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            attrs['project_id'] = project_id
        if parsed_args.description:
            attrs['description'] = parsed_args.description

        if parsed_args.name:
            attrs['name'] = str(parsed_args.name)
        attrs['type'] = parsed_args.type
        if parsed_args.type == 'subnet':
            _subnet_ids = [
                client.find_subnet(endpoint, ignore_missing=False)['id']
                for endpoint in parsed_args.endpoints
            ]
            attrs['endpoints'] = _subnet_ids
        else:
            attrs['endpoints'] = parsed_args.endpoints
        obj = client.create_vpn_endpoint_group(**attrs)
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj, _attr_map_dict, ['location', 'tenant_id']
        )
        data = utils.get_dict_properties(obj, columns)
        return display_columns, data


class DeleteEndpointGroup(command.Command):
    _description = _("Delete endpoint group(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'endpoint_group',
            metavar='<endpoint-group>',
            nargs='+',
            help=_('Endpoint group(s) to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0
        for endpoint in parsed_args.endpoint_group:
            try:
                endpoint_id = client.find_vpn_endpoint_group(
                    endpoint, ignore_missing=False
                ).id
                client.delete_vpn_endpoint_group(endpoint_id)
            except Exception as e:
                result += 1
                print(
                    f"Failed to delete endpoint group with "
                    f"name or ID {endpoint}: {e}"
                )

        if result > 0:
            total = len(parsed_args.endpoint_group)
            msg = _(
                "%(result)s of %(total)s endpoint group failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListEndpointGroup(command.Lister):
    _description = _("List endpoint groups that belong to a given project")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_("List additional fields in output"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network
        obj = client.vpn_endpoint_groups()
        headers, columns = column_util.get_column_definitions(
            _attr_map, long_listing=parsed_args.long
        )
        return (headers, (utils.get_dict_properties(s, columns) for s in obj))


class SetEndpointGroup(command.Command):
    _description = _("Set endpoint group properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description for the endpoint group'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Set a name for the endpoint group'),
        )
        parser.add_argument(
            'endpoint_group',
            metavar='<endpoint-group>',
            help=_('Endpoint group to set (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        attrs = {}
        if parsed_args.description:
            attrs['description'] = parsed_args.description
        if parsed_args.name:
            attrs['name'] = str(parsed_args.name)
        endpoint_id = client.find_vpn_endpoint_group(
            parsed_args.endpoint_group, ignore_missing=False
        )['id']
        try:
            client.update_vpn_endpoint_group(endpoint_id, **attrs)
        except Exception as e:
            msg = _(
                "Failed to set endpoint group %(endpoint_group)s: %(e)s"
            ) % {'endpoint_group': parsed_args.endpoint_group, 'e': e}
            raise exceptions.CommandError(msg)


class ShowEndpointGroup(command.ShowOne):
    _description = _("Display endpoint group details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'endpoint_group',
            metavar='<endpoint-group>',
            help=_('Endpoint group to display (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_vpn_endpoint_group(
            parsed_args.endpoint_group, ignore_missing=False
        )
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj, _attr_map_dict, ['location', 'tenant_id']
        )
        data = utils.get_dict_properties(obj, columns)
        return (display_columns, data)



#   Copyright 2014 CERN.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from osc_lib import exceptions

from openstack.identity.v3 import mapping as _mapping
from openstack.test import fakes as sdk_fakes
from openstackclient.identity.v3 import mapping
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


MAPPING_RULES_FILE_PATH = '/tmp/path/to/file'
# Copied from
# https://opendev.org/openstack/keystone/src/branch/stable/2026.1/keystone/tests/unit/mapping_fixtures.py
EMPLOYEE_GROUP_ID = "0cd5e9"
CONTRACTOR_GROUP_ID = "85a868"
TESTER_GROUP_ID = "123"
DEVELOPER_GROUP_ID = "xyz"
MAPPING_RULES = [
    {
        "local": [
            {"group": {"id": EMPLOYEE_GROUP_ID}},
            {"user": {"name": "{0}"}},
        ],
        "remote": [
            {"type": "UserName"},
            {
                "type": "orgPersonType",
                "not_any_of": ["Contractor", "SubContractor"],
            },
            {"type": "LastName", "any_one_of": ["Bo"]},
        ],
    },
    {
        "local": [
            {"group": {"id": CONTRACTOR_GROUP_ID}},
            {"user": {"name": "{0}"}},
        ],
        "remote": [
            {"type": "UserName"},
            {
                "type": "orgPersonType",
                "any_one_of": ["Contractor", "SubContractor"],
            },
            {"type": "FirstName", "any_one_of": ["Jill"]},
        ],
    },
]
MAPPING_RULES_2 = {
    "rules": [
        {
            "local": [
                {
                    "user": {"name": "{0} {1}", "email": "{2}"},
                    "group": {"id": EMPLOYEE_GROUP_ID},
                }
            ],
            "remote": [
                {"type": "FirstName"},
                {"type": "LastName"},
                {"type": "Email"},
                {
                    "type": "orgPersonType",
                    "any_one_of": ["Admin", "Big Cheese"],
                },
            ],
        },
        {
            "local": [{"user": {"name": "{0}", "email": "{1}"}}],
            "remote": [
                {"type": "UserName"},
                {"type": "Email"},
                {
                    "type": "orgPersonType",
                    "not_any_of": [
                        "Admin",
                        "Employee",
                        "Contractor",
                        "Tester",
                    ],
                },
            ],
        },
        {
            "local": [
                {"group": {"id": TESTER_GROUP_ID}},
                {"group": {"id": DEVELOPER_GROUP_ID}},
                {"user": {"name": "{0}"}},
            ],
            "remote": [
                {"type": "UserName"},
                {"type": "orgPersonType", "any_one_of": ["Tester"]},
                {
                    "type": "Email",
                    "any_one_of": [".*@example.com$"],
                    "regex": True,
                },
            ],
        },
    ]
}


class TestMappingCreate(identity_fakes.TestFederatedIdentity):
    def setUp(self):
        super().setUp()

        self.mapping = sdk_fakes.generate_fake_resource(
            _mapping.Mapping, rules=MAPPING_RULES, schema_version=None
        )
        self.identity_sdk_client.create_mapping.return_value = self.mapping

        self.cmd = mapping.CreateMapping(self.app, None)

    def test_create_mapping(self):
        arglist = [
            '--rules',
            MAPPING_RULES_FILE_PATH,
            self.mapping.id,
        ]
        verifylist = [
            ('mapping', self.mapping.id),
            ('rules', MAPPING_RULES_FILE_PATH),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        mocker = mock.Mock()
        mocker.return_value = MAPPING_RULES
        with mock.patch(
            "openstackclient.identity.v3.mapping.CreateMapping._read_rules",
            mocker,
        ):
            columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.create_mapping.assert_called_with(
            id=self.mapping.id, rules=MAPPING_RULES, schema_version=None
        )

        collist = ('id', 'rules', 'schema_version')
        self.assertEqual(collist, columns)

        datalist = (self.mapping.id, MAPPING_RULES, None)
        self.assertEqual(datalist, data)


class TestMappingDelete(identity_fakes.TestFederatedIdentity):
    def setUp(self):
        super().setUp()

        self.mapping = sdk_fakes.generate_fake_resource(
            _mapping.Mapping, rules=MAPPING_RULES, schema_version=None
        )
        self.identity_sdk_client.delete_mapping.return_value = None
        self.cmd = mapping.DeleteMapping(self.app, None)

    def test_delete_mapping(self):
        arglist = [self.mapping.id]
        verifylist = [('mapping', [self.mapping.id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_mapping.assert_called_with(
            self.mapping.id, ignore_missing=False
        )
        self.assertIsNone(result)


class TestMappingList(identity_fakes.TestFederatedIdentity):
    def setUp(self):
        super().setUp()

        self.mapping = sdk_fakes.generate_fake_resource(
            _mapping.Mapping, rules=MAPPING_RULES, schema_version='1.0'
        )

        self.extra_mapping = sdk_fakes.generate_fake_resource(
            _mapping.Mapping, rules=MAPPING_RULES_2, schema_version='2.0'
        )

        self.identity_sdk_client.mappings.return_value = [
            self.mapping,
            self.extra_mapping,
        ]

        # Get the command object to test
        self.cmd = mapping.ListMapping(self.app, None)

    def test_mapping_list(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.mappings.assert_called_with()

        collist = ('ID', 'schema_version')
        self.assertEqual(collist, columns)

        datalist = [
            (self.mapping.id, '1.0'),
            (self.extra_mapping.id, '2.0'),
        ]
        self.assertEqual(datalist, data)


class TestMappingSet(identity_fakes.TestFederatedIdentity):
    def setUp(self):
        super().setUp()

        self.mapping = sdk_fakes.generate_fake_resource(
            _mapping.Mapping, rules=MAPPING_RULES, schema_version='1.0'
        )
        self.mapping_2 = sdk_fakes.generate_fake_resource(
            _mapping.Mapping, rules=MAPPING_RULES_2, schema_version='2.0'
        )
        self.identity_sdk_client.get_mapping.return_value = self.mapping
        self.identity_sdk_client.update_mapping.return_value = self.mapping_2

        # Get the command object to test
        self.cmd = mapping.SetMapping(self.app, None)

    def test_set_new_rules(self):
        arglist = [
            '--rules',
            MAPPING_RULES_FILE_PATH,
            self.mapping.id,
        ]
        verifylist = [
            ('mapping', self.mapping.id),
            ('rules', MAPPING_RULES_FILE_PATH),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        mocker = mock.Mock()
        mocker.return_value = MAPPING_RULES_2
        with mock.patch(
            "openstackclient.identity.v3.mapping.SetMapping._read_rules",
            mocker,
        ):
            result = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.update_mapping.assert_called_with(
            mapping=self.mapping.id,
            rules=MAPPING_RULES_2,
            schema_version=None,
        )

        self.assertIsNone(result)

    def test_set_rules_wrong_file_path(self):
        arglist = [
            '--rules',
            MAPPING_RULES_FILE_PATH,
            self.mapping.id,
        ]
        verifylist = [
            ('mapping', self.mapping.id),
            ('rules', MAPPING_RULES_FILE_PATH),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestMappingShow(identity_fakes.TestFederatedIdentity):
    def setUp(self):
        super().setUp()

        self.mapping = sdk_fakes.generate_fake_resource(
            _mapping.Mapping, rules=MAPPING_RULES, schema_version=None
        )
        self.identity_sdk_client.get_mapping.return_value = self.mapping

        self.cmd = mapping.ShowMapping(self.app, None)

    def test_mapping_show(self):
        arglist = [self.mapping.id]
        verifylist = [('mapping', self.mapping.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.get_mapping.assert_called_with(
            self.mapping.id
        )

        collist = ('id', 'rules', 'schema_version')
        self.assertEqual(collist, columns)

        datalist = (self.mapping.id, MAPPING_RULES, None)
        self.assertEqual(datalist, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Compute v2 API Library Tests"""

import http
from unittest import mock
import uuid

from openstack.compute import v2 as compute_v2
from osc_lib import exceptions as osc_lib_exceptions

from openstackclient.api import compute_v2 as compute
from openstackclient.tests.unit import fakes
from openstackclient.tests.unit import utils


class TestSecurityGroup(utils.TestCase):
    def setUp(self):
        super().setUp()

        # TODO(stephenfin): Switch to spec_set once keystoneauth exposes
        # instance attributes as class attributes
        # https://review.opendev.org/c/openstack/keystoneauth/+/994090
        self.compute_client = mock.Mock(spec=compute_v2.Proxy)

    def test_create_security_group(self):
        sg_name = 'name-' + uuid.uuid4().hex
        sg_description = 'description-' + uuid.uuid4().hex
        data = {
            'security_group': {
                'id': uuid.uuid4().hex,
                'name': sg_name,
                'description': sg_description,
                'tenant_id': 'project-id-' + uuid.uuid4().hex,
                'rules': [],
            }
        }
        self.compute_client.post.return_value = fakes.FakeResponse(data=data)

        result = compute.create_security_group(
            self.compute_client, sg_name, sg_description
        )

        self.compute_client.post.assert_called_once_with(
            '/os-security-groups',
            data={'name': sg_name, 'description': sg_description},
            microversion='2.1',
        )
        self.assertEqual(data['security_group'], result)

    def test_list_security_groups(self):
        data = {
            'security_groups': [
                {
                    'id': uuid.uuid4().hex,
                    'name': uuid.uuid4().hex,
                    'description': 'description-' + uuid.uuid4().hex,
                    'tenant_id': 'project-id-' + uuid.uuid4().hex,
                    'rules': [],
                }
            ],
        }
        self.compute_client.get.return_value = fakes.FakeResponse(data=data)

        result = compute.list_security_groups(self.compute_client)

        self.compute_client.get.assert_called_once_with(
            '/os-security-groups', microversion='2.1'
        )
        self.assertEqual(data['security_groups'], result)

    def test_find_security_group_by_id(self):
        sg_id = uuid.uuid4().hex
        sg_name = 'name-' + uuid.uuid4().hex
        data = {
            'security_group': {
                'id': sg_id,
                'name': sg_name,
                'description': 'description-' + uuid.uuid4().hex,
                'tenant_id': 'project-id-' + uuid.uuid4().hex,
                'rules': [],
            }
        }
        self.compute_client.get.side_effect = [
            fakes.FakeResponse(data=data),
        ]

        result = compute.find_security_group(self.compute_client, sg_id)

        self.compute_client.get.assert_has_calls(
            [
                mock.call(f'/os-security-groups/{sg_id}', microversion='2.1'),
            ]
        )
        self.assertEqual(data['security_group'], result)

    def test_find_security_group_by_name(self):
        sg_id = uuid.uuid4().hex
        sg_name = 'name-' + uuid.uuid4().hex
        data = {
            'security_groups': [
                {
                    'id': sg_id,
                    'name': sg_name,
                    'description': 'description-' + uuid.uuid4().hex,
                    'tenant_id': 'project-id-' + uuid.uuid4().hex,
                    'rules': [],
                }
            ],
        }
        self.compute_client.get.side_effect = [
            fakes.FakeResponse(status_code=http.HTTPStatus.NOT_FOUND),
            fakes.FakeResponse(data=data),
        ]

        result = compute.find_security_group(self.compute_client, sg_name)

        self.compute_client.get.assert_has_calls(
            [
                mock.call(
                    f'/os-security-groups/{sg_name}', microversion='2.1'
                ),
                mock.call('/os-security-groups', microversion='2.1'),
            ]
        )
        self.assertEqual(data['security_groups'][0], result)

    def test_find_security_group_not_found(self):
        data = {'security_groups': []}
        self.compute_client.get.side_effect = [
            fakes.FakeResponse(status_code=http.HTTPStatus.NOT_FOUND),
            fakes.FakeResponse(data=data),
        ]
        self.assertRaises(
            osc_lib_exceptions.NotFound,
            compute.find_security_group,
            self.compute_client,
            'invalid-sg',
        )

    def test_find_security_group_by_name_duplicate(self):
        sg_name = 'name-' + uuid.uuid4().hex
        data = {
            'security_groups': [
                {
                    'id': uuid.uuid4().hex,
                    'name': sg_name,
                    'description': 'description-' + uuid.uuid4().hex,
                    'tenant_id': 'project-id-' + uuid.uuid4().hex,
                    'rules': [],
                },
                {
                    'id': uuid.uuid4().hex,
                    'name': sg_name,
                    'description': 'description-' + uuid.uuid4().hex,
                    'tenant_id': 'project-id-' + uuid.uuid4().hex,
                    'rules': [],
                },
            ],
        }
        self.compute_client.get.side_effect = [
            fakes.FakeResponse(status_code=http.HTTPStatus.NOT_FOUND),
            fakes.FakeResponse(data=data),
        ]

        self.assertRaises(
            osc_lib_exceptions.NotFound,
            compute.find_security_group,
            self.compute_client,
            sg_name,
        )

    def test_update_security_group(self):
        sg_id = uuid.uuid4().hex
        sg_name = 'name-' + uuid.uuid4().hex
        sg_description = 'description-' + uuid.uuid4().hex
        data = {
            'security_group': {
                'id': sg_id,
                'name': sg_name,
                'description': sg_description,
                'tenant_id': 'project-id-' + uuid.uuid4().hex,
                'rules': [],
            }
        }
        self.compute_client.put.return_value = fakes.FakeResponse(data=data)

        result = compute.update_security_group(
            self.compute_client, sg_id, sg_name, sg_description
        )

        self.compute_client.put.assert_called_once_with(
            f'/os-security-groups/{sg_id}',
            data={'name': sg_name, 'description': sg_description},
            microversion='2.1',
        )
        self.assertEqual(data['security_group'], result)

    def test_delete_security_group(self):
        sg_id = uuid.uuid4().hex
        self.compute_client.delete.return_value = fakes.FakeResponse(
            status_code=http.HTTPStatus.NO_CONTENT
        )

        result = compute.delete_security_group(self.compute_client, sg_id)

        self.compute_client.delete.assert_called_once_with(
            f'/os-security-groups/{sg_id}',
            microversion='2.1',
        )
        self.assertIsNone(result)


class TestSecurityGroupRule(utils.TestCase):
    def setUp(self):
        super().setUp()

        self.compute_client = mock.Mock(spec=compute_v2.Proxy)

    def test_create_security_group_rule(self):
        sg_id = uuid.uuid4().hex
        data = {
            'security_group_rule': {
                'parent_group_id': sg_id,
                'ip_protocol': 'tcp',
                'from_port': 22,
                'to_port': 22,
                'cidr': '10.0.0.0/24',
            }
        }
        self.compute_client.post.return_value = fakes.FakeResponse(data=data)

        result = compute.create_security_group_rule(
            self.compute_client,
            security_group_id=sg_id,
            ip_protocol='tcp',
            from_port=22,
            to_port=22,
            remote_ip='10.0.0.0/24',
            remote_group=None,
        )

        self.compute_client.post.assert_called_once_with(
            '/os-security-group-rules',
            data={
                'parent_group_id': sg_id,
                'ip_protocol': 'tcp',
                'from_port': 22,
                'to_port': 22,
                'cidr': '10.0.0.0/24',
                'group_id': None,
            },
            microversion='2.1',
        )
        self.assertEqual(data['security_group_rule'], result)

    def test_delete_security_group_rule(self):
        sg_id = uuid.uuid4().hex
        self.compute_client.delete.return_value = fakes.FakeResponse(
            status_code=http.HTTPStatus.NO_CONTENT
        )

        result = compute.delete_security_group_rule(self.compute_client, sg_id)

        self.compute_client.delete.assert_called_once_with(
            f'/os-security-group-rules/{sg_id}',
            microversion='2.1',
        )
        self.assertIsNone(result)


class TestNetwork(utils.TestCase):
    def setUp(self):
        super().setUp()

        self.compute_client = mock.Mock(spec=compute_v2.Proxy)

    def test_create_network(self):
        net_name = 'name-' + uuid.uuid4().hex
        net_subnet = '10.0.0.0/24'
        data = {
            'network': {
                'id': uuid.uuid4().hex,
                'label': net_name,
                'cidr': net_subnet,
                'share_address': True,
                # other fields omitted for brevity
            }
        }
        self.compute_client.post.return_value = fakes.FakeResponse(data=data)

        result = compute.create_network(
            self.compute_client,
            name=net_name,
            subnet=net_subnet,
            share_subnet=True,
        )

        self.compute_client.post.assert_called_once_with(
            '/os-networks',
            data={
                'label': net_name,
                'cidr': net_subnet,
                'share_address': True,
            },
            microversion='2.1',
        )
        self.assertEqual(data['network'], result)

    def test_list_networks(self):
        data = {
            'networks': [
                {
                    'id': uuid.uuid4().hex,
                    'label': f'name-{uuid.uuid4().hex}',
                    # other fields omitted for brevity
                }
            ],
        }
        self.compute_client.get.return_value = fakes.FakeResponse(data=data)

        result = compute.list_networks(self.compute_client)

        self.compute_client.get.assert_called_once_with(
            '/os-networks', microversion='2.1'
        )
        self.assertEqual(data['networks'], result)

    def test_find_network_by_id(self):
        net_id = uuid.uuid4().hex
        net_name = 'name-' + uuid.uuid4().hex
        data = {
            'network': {
                'id': net_id,
                'label': net_name,
                # other fields omitted for brevity
            }
        }
        self.compute_client.get.side_effect = [
            fakes.FakeResponse(data=data),
        ]

        result = compute.find_network(self.compute_client, net_id)

        self.compute_client.get.assert_has_calls(
            [
                mock.call(f'/os-networks/{net_id}', microversion='2.1'),
            ]
        )
        self.assertEqual(data['network'], result)

    def test_find_network_by_name(self):
        net_id = uuid.uuid4().hex
        net_name = 'name-' + uuid.uuid4().hex
        data = {
            'networks': [
                {
                    'id': net_id,
                    'label': net_name,
                    # other fields omitted for brevity
                }
            ],
        }
        self.compute_client.get.side_effect = [
            fakes.FakeResponse(status_code=http.HTTPStatus.NOT_FOUND),
            fakes.FakeResponse(data=data),
        ]

        result = compute.find_network(self.compute_client, net_name)

        self.compute_client.get.assert_has_calls(
            [
                mock.call(f'/os-networks/{net_name}', microversion='2.1'),
                mock.call('/os-networks', microversion='2.1'),
            ]
        )
        self.assertEqual(data['networks'][0], result)

    def test_find_network_not_found(self):
        data = {'networks': []}
        self.compute_client.get.side_effect = [
            fakes.FakeResponse(status_code=http.HTTPStatus.NOT_FOUND),
            fakes.FakeResponse(data=data),
        ]
        self.assertRaises(
            osc_lib_exceptions.NotFound,
            compute.find_network,
            self.compute_client,
            'invalid-net',
        )

    def test_find_network_by_name_duplicate(self):
        net_name = 'name-' + uuid.uuid4().hex
        data = {
            'networks': [
                {
                    'id': uuid.uuid4().hex,
                    'label': net_name,
                    # other fields omitted for brevity
                },
                {
                    'id': uuid.uuid4().hex,
                    'label': net_name,
                    # other fields omitted for brevity
                },
            ],
        }
        self.compute_client.get.side_effect = [
            fakes.FakeResponse(status_code=http.HTTPStatus.NOT_FOUND),
            fakes.FakeResponse(data=data),
        ]

        self.assertRaises(
            osc_lib_exceptions.NotFound,
            compute.find_network,
            self.compute_client,
            net_name,
        )

    def test_delete_network(self):
        net_id = uuid.uuid4().hex
        self.compute_client.delete.return_value = fakes.FakeResponse(
            status_code=http.HTTPStatus.NO_CONTENT
        )

        result = compute.delete_network(self.compute_client, net_id)

        self.compute_client.delete.assert_called_once_with(
            f'/os-networks/{net_id}', microversion='2.1'
        )
        self.assertIsNone(result)


class TestFloatingIP(utils.TestCase):
    def setUp(self):
        super().setUp()

        self.compute_client = mock.Mock(spec=compute_v2.Proxy)

    def test_create_floating_ip(self):
        network = 'network-' + uuid.uuid4().hex
        data = {
            'floating_ip': {
                'fixed_ip': None,
                'id': uuid.uuid4().hex,
                'instance_id': None,
                'ip': '172.24.4.17',
                'pool': network,
            }
        }
        self.compute_client.post.return_value = fakes.FakeResponse(data=data)

        result = compute.create_floating_ip(
            self.compute_client, network=network
        )

        self.compute_client.post.assert_called_once_with(
            '/os-floating-ips', data={'pool': network}, microversion='2.1'
        )
        self.assertEqual(data['floating_ip'], result)

    def test_list_floating_ips(self):
        data = {
            'floating_ips': [
                {
                    'fixed_ip': None,
                    'id': uuid.uuid4().hex,
                    'instance_id': None,
                    'ip': '172.24.4.17',
                    'pool': f'network-{uuid.uuid4().hex}',
                }
            ],
        }
        self.compute_client.get.return_value = fakes.FakeResponse(data=data)

        result = compute.list_floating_ips(self.compute_client)

        self.compute_client.get.assert_called_once_with(
            '/os-floating-ips', microversion='2.1'
        )
        self.assertEqual(data['floating_ips'], result)

    def test_get_floating_ip(self):
        fip_id = uuid.uuid4().hex
        data = {
            'floating_ip': {
                'fixed_ip': None,
                'id': fip_id,
                'instance_id': None,
                'ip': '172.24.4.17',
                'pool': f'network-{uuid.uuid4().hex}',
            }
        }
        self.compute_client.get.side_effect = [
            fakes.FakeResponse(data=data),
        ]

        result = compute.get_floating_ip(self.compute_client, fip_id)

        self.compute_client.get.assert_called_once_with(
            f'/os-floating-ips/{fip_id}', microversion='2.1'
        )
        self.assertEqual(data['floating_ip'], result)

    def test_delete_floating_ip(self):
        fip_id = uuid.uuid4().hex
        self.compute_client.delete.return_value = fakes.FakeResponse(
            status_code=http.HTTPStatus.NO_CONTENT
        )

        result = compute.delete_floating_ip(self.compute_client, fip_id)

        self.compute_client.delete.assert_called_once_with(
            f'/os-floating-ips/{fip_id}', microversion='2.1'
        )
        self.assertIsNone(result)


class TestFloatingIPPool(utils.TestCase):
    def setUp(self):
        super().setUp()

        self.compute_client = mock.Mock(spec=compute_v2.Proxy)

    def test_list_floating_ip_pools(self):
        data = {
            'floating_ip_pools': [
                {
                    'name': f'pool-{uuid.uuid4().hex}',
                }
            ],
        }
        self.compute_client.get.return_value = fakes.FakeResponse(data=data)

        result = compute.list_floating_ip_pools(self.compute_client)

        self.compute_client.get.assert_called_once_with(
            '/os-floating-ip-pools', microversion='2.1'
        )
        self.assertEqual(data['floating_ip_pools'], result)



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Manage access to the clients, including authenticating when needed."""

import argparse
from collections.abc import Callable
import importlib
import logging
import sys
from typing import Any, Protocol, TypeVar, runtime_checkable

from keystoneauth1 import access as ksa_access
from openstack.block_storage import v2 as volume_v2
from openstack.block_storage import v3 as volume_v3
from openstack.compute import v2 as compute_v2
from openstack.image import v2 as image_v2
from openstack.network import v2 as network_v2
from osc_lib.cli import client_config
from osc_lib import clientmanager
from osc_lib import shell
import stevedore

from openstackclient.api import object_store_v1

LOG = logging.getLogger(__name__)
PLUGIN_MODULES: list[Any] = []
USER_AGENT = 'python-openstackclient'


class ClientManager(clientmanager.ClientManager):
    """Manages access to API clients, including authentication

    Wrap osc_lib's ClientManager to maintain compatibility for the existing
    plugin V2 interface.  Some currently private attributes become public
    in osc-lib so we need to maintain a transition period.
    """

    # we know this will be set by us and will not be nullable
    auth_ref: ksa_access.AccessInfo

    # this is a hack to keep mypy happy: the actual attributes are set in
    # get_plugin_modules below
    # TODO(stephenfin): Change the types of identity, object store and share
    # once we've migrated everything to SDK
    compute: compute_v2.Proxy
    identity: Any
    image: image_v2.Proxy
    network: network_v2.Proxy
    object_store: object_store_v1.APIv1
    share: Any
    volume: volume_v2.Proxy | volume_v3.Proxy

    def __init__(
        self,
        cli_options: Any = None,
        api_version: Any = None,
        pw_func: Any = None,
    ) -> None:
        super().__init__(
            cli_options=cli_options,
            api_version=api_version,
            pw_func=pw_func,
        )

        # TODO(dtroyer): For compatibility; mark this for removal when plugin
        # interface v2 is removed
        self._region_name = self.region_name
        self._interface = self.interface
        self._cacert = self.cacert
        self._insecure = not self.verify
        # store original auth_type
        self._original_auth_type = cli_options.auth_type

    def setup_auth(self) -> None:
        """Set up authentication"""

        if self._auth_setup_completed:
            return

        # NOTE(dtroyer): Validate the auth args; this is protected with 'if'
        #                because openstack_config is an optional argument to
        #                CloudConfig.__init__() and we'll die if it was not
        #                passed.
        if (
            self._auth_required
            and self._cli_options._openstack_config is not None
        ):
            if not isinstance(
                self._cli_options._openstack_config, client_config.OSC_Config
            ):
                # programmer error
                raise TypeError('unexpected type for _openstack_config')

            self._cli_options._openstack_config._pw_callback = (
                shell.prompt_for_password
            )
            try:
                # We might already get auth from SDK caching
                if not self._cli_options._auth:
                    self._cli_options._auth = (
                        self._cli_options._openstack_config.load_auth_plugin(
                            self._cli_options.config,
                        )
                    )
            except TypeError as e:
                self._fallback_load_auth_plugin(e)

        return super().setup_auth()

    def _fallback_load_auth_plugin(self, e: Exception) -> None:
        # NOTES(RuiChen): Hack to avoid auth plugins choking on data they don't
        #                 expect, delete fake token and endpoint, then try to
        #                 load auth plugin again with user specified options.
        #                 We know it looks ugly, but it's necessary.
        if self._cli_options.config['auth']['token'] == 'x':  # noqa: S105
            # restore original auth_type
            self._cli_options.config['auth_type'] = self._original_auth_type
            del self._cli_options.config['auth']['token']
            del self._cli_options.config['auth']['endpoint']

            if not isinstance(
                self._cli_options._openstack_config, client_config.OSC_Config
            ):
                # programmer error
                raise TypeError('unexpected type for _openstack_config')

            self._cli_options._auth = (
                self._cli_options._openstack_config.load_auth_plugin(
                    self._cli_options.config,
                )
            )
        else:
            raise e

    def is_network_endpoint_enabled(self) -> bool:
        """Check if the network endpoint is enabled"""
        # NOTE(dtroyer): is_service_available() can also return None if
        #                there is no Service Catalog, callers here are
        #                not expecting that so fold None into True to
        #                use Network API by default
        return (
            self.is_service_available(
                'network', self.region_name, self.interface
            )
            is not False
        )

    def is_compute_endpoint_enabled(self) -> bool:
        """Check if Compute endpoint is enabled"""
        return (
            self.is_service_available(
                'compute', self.region_name, self.interface
            )
            is not False
        )

    # TODO(stephenfin): Drop volume_client argument in OSC 8.0 or later.
    def is_volume_endpoint_enabled(self, volume_client: Any = None) -> bool:
        """Check if volume endpoint is enabled"""
        # We check against the service type and all aliases defined by the
        # Service Types Authority
        # https://service-types.openstack.org/service-types.json
        return (
            self.is_service_available(
                'block-storage', self.region_name, self.interface
            )
            is not False
            or self.is_service_available(
                'volume', self.region_name, self.interface
            )
            is not False
            or self.is_service_available(
                'volumev3', self.region_name, self.interface
            )
            is not False
            or self.is_service_available(
                'volumev2', self.region_name, self.interface
            )
            is not False
            or self.is_service_available(
                'block-store', self.region_name, self.interface
            )
            is not False
        )


# Plugin Support

ArgumentParserT = TypeVar('ArgumentParserT', bound=argparse.ArgumentParser)


@runtime_checkable  # Optional: allows usage with isinstance()
class PluginModule(Protocol):
    DEFAULT_API_VERSION: str
    API_VERSION_OPTION: str
    API_NAME: str
    API_VERSIONS: tuple[str]

    make_client: Callable[..., Any]
    build_option_parser: Callable[[ArgumentParserT], ArgumentParserT]
    check_api_version: Callable[[str], bool]


def _on_load_failure_callback(
    manager: stevedore.ExtensionManager[PluginModule],
    ep: importlib.metadata.EntryPoint,
    err: BaseException,
) -> None:
    sys.stderr.write(
        f"WARNING: Failed to import plugin {ep.group}:{ep.name}: {err}.\n"
    )


def get_plugin_modules(group: str) -> list[Any]:
    """Find plugin entry points"""
    mod_list = []
    mgr: stevedore.ExtensionManager[PluginModule]
    mgr = stevedore.ExtensionManager(
        group, on_load_failure_callback=_on_load_failure_callback
    )
    for ep in mgr:
        LOG.debug('Found plugin %s', ep.name)

        module_name = ep.entry_point.module

        try:
            module = importlib.import_module(module_name)
        except Exception as err:
            sys.stderr.write(
                f"WARNING: Failed to import plugin "
                f"{ep.module_name}:{ep.name}: {err}.\n"
            )
            continue

        mod_list.append(module)

        # Add the plugin to the ClientManager
        setattr(
            clientmanager.ClientManager,
            module.API_NAME,
            clientmanager.ClientCache(
                getattr(sys.modules[module_name], 'make_client', None)
            ),
        )
    return mod_list


def build_plugin_option_parser(
    parser: ArgumentParserT,
) -> ArgumentParserT:
    """Add plugin options to the parser"""

    # Loop through extensions to get parser additions
    for mod in PLUGIN_MODULES:
        parser = mod.build_option_parser(parser)
    return parser


# Get list of base plugin modules
PLUGIN_MODULES = get_plugin_modules(
    'openstack.cli.base',
)
# Append list of external plugin modules
PLUGIN_MODULES.extend(
    get_plugin_modules(
        'openstack.cli.extension',
    )
)



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Module action implementation"""

import argparse
import importlib.metadata
import sys
from collections.abc import Iterable, Sequence
from typing import Any

from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


class ListCommand(command.Lister):
    _description = _("List recognized commands by group")

    auth_required = False

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--group',
            metavar='<group-keyword>',
            help=_(
                'Show commands filtered by a command group, for example: '
                'identity, volume, compute, image, network and '
                'other keywords'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        cm = self.app.command_manager
        groups = cm.get_command_groups()
        groups = sorted(groups)
        columns = ('Command Group', 'Commands')

        if parsed_args.group:
            groups = sorted(
                group for group in groups if parsed_args.group in group
            )

        commands = []
        for group in groups:
            command_names = cm.get_command_names(group)
            command_names = sorted(command_names)

            if command_names != []:
                # TODO(bapalm): Fix this when cliff properly supports
                # handling the detection rather than using the hard-code below.
                if parsed_args.formatter == 'table':
                    command_names = utils.format_list(command_names, "\n")  # type: ignore[assignment]

                commands.append((group, command_names))

        return (columns, commands)


class ListModule(command.ShowOne):
    _description = _("List module versions")

    auth_required = False

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all',
            action='store_true',
            default=False,
            help=_('Show all modules that have version information'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        data = {}
        # Get module versions
        package_distributions = importlib.metadata.packages_distributions()
        for k in sys.modules.keys():
            module = k.split('.')[0]
            # Skip private modules and the modules that had been added,
            # like: keystoneclient, keystoneclient.exceptions and
            # keystoneclient.auth
            if module.startswith('_') or module in data:
                continue

            # TODO(dtroyer): Need a better way to decide which modules to
            #                show for the default (not --all) invocation.
            #                It should be just the things we actually care
            #                about like client and plugin modules...
            # Handle xxxclient and openstacksdk
            if parsed_args.all or (
                module.endswith('client') or module == 'openstack'
            ):
                try:
                    package = package_distributions[module][0]
                    data[module] = importlib.metadata.version(package)
                except Exception:  # noqa: S110
                    # Catch all exceptions, just skip it
                    pass

        col_headers, col_data = zip(*sorted(data.items()))
        return col_headers, col_data



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Volume v3 host action implementations"""

import argparse

from openstack.block_storage.v3 import service as _service
from openstack import utils as sdk_utils

from openstackclient import command
from openstackclient.i18n import _


class FailoverVolumeHost(command.Command):
    _description = _("Failover volume host to different backend")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "host", metavar="<host-name>", help=_("Name of volume host")
        )
        parser.add_argument(
            "--volume-backend",
            metavar="<backend-id>",
            required=True,
            help=_(
                "The ID of the volume backend replication "
                "target where the host will failover to (required)"
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        service = _service.Service(host=parsed_args.host)
        volume_client.failover_service(
            service, backend_id=parsed_args.volume_backend
        )


class SetVolumeHost(command.Command):
    _description = _("Set volume host properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "host", metavar="<host-name>", help=_("Name of volume host")
        )
        enabled_group = parser.add_mutually_exclusive_group()
        enabled_group.add_argument(
            "--disable",
            action="store_true",
            help=_("Freeze and disable the specified volume host"),
        )
        enabled_group.add_argument(
            "--enable",
            action="store_true",
            help=_("Thaw and enable the specified volume host"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        service = _service.Service(host=parsed_args.host)
        if parsed_args.enable:
            volume_client.thaw_service(service)
        if parsed_args.disable:
            volume_client.freeze_service(service)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstack import exceptions as sdk_exc
from openstack.identity.v3 import region as _region
from openstack.identity.v3 import registered_limit as _registered_limit
from openstack.identity.v3 import service as _service
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.identity.v3 import registered_limit
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestRegisteredLimitCreate(identity_fakes.TestIdentity):
    def setUp(self):
        super().setUp()

        self.service = sdk_fakes.generate_fake_resource(_service.Service)
        self.region = sdk_fakes.generate_fake_resource(_region.Region)

        self.description = 'default limit of foobars'
        self.default_limit = 10
        self.resource_name = 'foobars'

        self.identity_sdk_client.find_service.return_value = self.service
        self.identity_sdk_client.get_region.return_value = self.region

        self.registered_limit = sdk_fakes.generate_fake_resource(
            resource_type=_registered_limit.RegisteredLimit,
            description=None,
            region_id=None,
            service_id=self.service.id,
            default_limit=self.default_limit,
            resource_name=self.resource_name,
        )
        self.registered_limit_with_options = sdk_fakes.generate_fake_resource(
            resource_type=_registered_limit.RegisteredLimit,
            description=self.description,
            region_id=self.region.id,
            service_id=self.service.id,
            default_limit=self.default_limit,
            resource_name=self.resource_name,
        )

        self.cmd = registered_limit.CreateRegisteredLimit(self.app, None)

    def test_registered_limit_create_without_options(self):
        self.identity_sdk_client.create_registered_limit.return_value = (
            self.registered_limit
        )

        arglist = [
            '--service',
            self.service.id,
            '--default-limit',
            str(self.default_limit),
            self.resource_name,
        ]

        verifylist = [
            ('service', self.service.id),
            ('default_limit', self.default_limit),
            ('resource_name', self.resource_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'service_id': self.service.id,
            'default_limit': self.default_limit,
            'resource_name': self.resource_name,
        }
        self.identity_sdk_client.create_registered_limit.assert_called_with(
            **kwargs
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )

        self.assertEqual(collist, columns)
        datalist = (
            self.default_limit,
            None,
            self.registered_limit.id,
            None,
            self.resource_name,
            self.service.id,
        )
        self.assertEqual(datalist, data)

    def test_registered_limit_create_with_options(self):
        self.identity_sdk_client.create_registered_limit.return_value = (
            self.registered_limit_with_options
        )

        arglist = [
            '--region',
            self.region.id,
            '--description',
            self.description,
            '--service',
            self.service.id,
            '--default-limit',
            str(self.default_limit),
            self.resource_name,
        ]

        verifylist = [
            ('region', self.region.id),
            ('description', self.description),
            ('service', self.service.id),
            ('default_limit', self.default_limit),
            ('resource_name', self.resource_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'description': self.description,
            'region_id': self.region.id,
            'service_id': self.service.id,
            'default_limit': self.default_limit,
            'resource_name': self.resource_name,
        }
        self.identity_sdk_client.create_registered_limit.assert_called_with(
            **kwargs
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )

        self.assertEqual(collist, columns)
        datalist = (
            self.default_limit,
            self.description,
            self.registered_limit_with_options.id,
            self.region.id,
            self.resource_name,
            self.service.id,
        )
        self.assertEqual(datalist, data)


class TestRegisteredLimitDelete(identity_fakes.TestIdentity):
    def setUp(self):
        super().setUp()

        self.cmd = registered_limit.DeleteRegisteredLimit(self.app, None)

    def test_registered_limit_delete(self):
        self.registered_limit = sdk_fakes.generate_fake_resource(
            resource_type=_registered_limit.RegisteredLimit,
        )
        self.identity_sdk_client.delete_registered_limit.return_value = None

        arglist = [self.registered_limit.id]
        verifylist = [('registered_limits', [self.registered_limit.id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_registered_limit.assert_called_with(
            self.registered_limit.id,
            ignore_missing=False,
        )
        self.assertIsNone(result)

    def test_registered_limit_delete_with_exception(self):
        self.identity_sdk_client.delete_registered_limit.side_effect = (
            sdk_exc.ResourceNotFound
        )

        arglist = ['fake-registered-limit-id']
        verifylist = [('registered_limits', ['fake-registered-limit-id'])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual(
                '1 of 1 registered limits failed to delete.', str(e)
            )


class TestRegisteredLimitShow(identity_fakes.TestIdentity):
    def setUp(self):
        super().setUp()

        self.service = sdk_fakes.generate_fake_resource(_service.Service)
        self.region = sdk_fakes.generate_fake_resource(_region.Region)

        self.description = 'default limit of foobars'
        self.default_limit = 10
        self.resource_name = 'foobars'

        self.identity_sdk_client.find_service.return_value = self.service
        self.identity_sdk_client.get_region.return_value = self.region

        self.registered_limit = sdk_fakes.generate_fake_resource(
            resource_type=_registered_limit.RegisteredLimit,
            description=None,
            region_id=None,
            service_id=self.service.id,
            default_limit=self.default_limit,
            resource_name=self.resource_name,
        )
        self.registered_limit_with_options = sdk_fakes.generate_fake_resource(
            resource_type=_registered_limit.RegisteredLimit,
            description=self.description,
            region_id=self.region.id,
            service_id=self.service.id,
            default_limit=self.default_limit,
            resource_name=self.resource_name,
        )

        self.cmd = registered_limit.ShowRegisteredLimit(self.app, None)

    def test_registered_limit_show(self):
        self.identity_sdk_client.get_registered_limit.return_value = (
            self.registered_limit
        )

        arglist = [self.registered_limit.id]
        verifylist = [('registered_limit_id', self.registered_limit.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.get_registered_limit.assert_called_with(
            self.registered_limit.id
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            self.default_limit,
            None,
            self.registered_limit.id,
            None,
            self.resource_name,
            self.service.id,
        )
        self.assertEqual(datalist, data)

    def test_registered_limit_show_with_options(self):
        self.identity_sdk_client.get_registered_limit.return_value = (
            self.registered_limit_with_options
        )

        arglist = [self.registered_limit_with_options.id]
        verifylist = [
            ('registered_limit_id', self.registered_limit_with_options.id)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.get_registered_limit.assert_called_with(
            self.registered_limit_with_options.id
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            self.default_limit,
            self.description,
            self.registered_limit_with_options.id,
            self.region.id,
            self.resource_name,
            self.service.id,
        )
        self.assertEqual(datalist, data)


class TestRegisteredLimitSet(identity_fakes.TestIdentity):
    def setUp(self):
        super().setUp()

        self.service = sdk_fakes.generate_fake_resource(_service.Service)
        self.region = sdk_fakes.generate_fake_resource(_region.Region)

        self.default_limit = 10
        self.resource_name = 'foobars'

        self.identity_sdk_client.find_service.return_value = self.service
        self.identity_sdk_client.get_region.return_value = self.region

        self.registered_limit = sdk_fakes.generate_fake_resource(
            resource_type=_registered_limit.RegisteredLimit,
            description=None,
            region_id=None,
            service_id=self.service.id,
            default_limit=self.default_limit,
            resource_name=self.resource_name,
        )

        self.cmd = registered_limit.SetRegisteredLimit(self.app, None)

    def test_registered_limit_set_description(self):
        updated_description = 'default limit of foobars'
        updated_registered_limit = sdk_fakes.generate_fake_resource(
            resource_type=_registered_limit.RegisteredLimit,
            id=self.registered_limit.id,
            description=updated_description,
            region_id=None,
            service_id=self.service.id,
            default_limit=self.default_limit,
            resource_name=self.resource_name,
        )
        self.identity_sdk_client.update_registered_limit.return_value = (
            updated_registered_limit
        )

        arglist = [
            '--description',
            updated_description,
            self.registered_limit.id,
        ]
        verifylist = [
            ('description', updated_description),
            ('registered_limit_id', self.registered_limit.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.update_registered_limit.assert_called_with(
            self.registered_limit.id,
            description=updated_description,
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            self.default_limit,
            updated_description,
            self.registered_limit.id,
            None,
            self.resource_name,
            self.service.id,
        )
        self.assertEqual(datalist, data)

    def test_registered_limit_set_default_limit(self):
        updated_default_limit = 20
        updated_registered_limit = sdk_fakes.generate_fake_resource(
            resource_type=_registered_limit.RegisteredLimit,
            id=self.registered_limit.id,
            description=None,
            region_id=None,
            service_id=self.service.id,
            default_limit=updated_default_limit,
            resource_name=self.resource_name,
        )
        self.identity_sdk_client.update_registered_limit.return_value = (
            updated_registered_limit
        )

        arglist = [
            '--default-limit',
            str(updated_default_limit),
            self.registered_limit.id,
        ]
        verifylist = [
            ('default_limit', updated_default_limit),
            ('registered_limit_id', self.registered_limit.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.update_registered_limit.assert_called_with(
            self.registered_limit.id,
            default_limit=updated_default_limit,
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            updated_default_limit,
            None,
            self.registered_limit.id,
            None,
            self.resource_name,
            self.service.id,
        )
        self.assertEqual(datalist, data)

    def test_registered_limit_set_resource_name(self):
        updated_resource_name = 'volumes'
        updated_registered_limit = sdk_fakes.generate_fake_resource(
            resource_type=_registered_limit.RegisteredLimit,
            id=self.registered_limit.id,
            description=None,
            region_id=None,
            service_id=self.service.id,
            default_limit=self.default_limit,
            resource_name=updated_resource_name,
        )
        self.identity_sdk_client.update_registered_limit.return_value = (
            updated_registered_limit
        )

        arglist = [
            '--resource-name',
            updated_resource_name,
            self.registered_limit.id,
        ]
        verifylist = [
            ('resource_name', updated_resource_name),
            ('registered_limit_id', self.registered_limit.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.update_registered_limit.assert_called_with(
            self.registered_limit.id,
            resource_name=updated_resource_name,
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            self.default_limit,
            None,
            self.registered_limit.id,
            None,
            updated_resource_name,
            self.service.id,
        )
        self.assertEqual(datalist, data)

    def test_registered_limit_set_service(self):
        updated_service = sdk_fakes.generate_fake_resource(_service.Service)
        self.identity_sdk_client.find_service.return_value = updated_service
        updated_registered_limit = sdk_fakes.generate_fake_resource(
            resource_type=_registered_limit.RegisteredLimit,
            id=self.registered_limit.id,
            description=None,
            region_id=None,
            service_id=updated_service.id,
            default_limit=self.default_limit,
            resource_name=self.resource_name,
        )
        self.identity_sdk_client.update_registered_limit.return_value = (
            updated_registered_limit
        )

        arglist = ['--service', updated_service.id, self.registered_limit.id]
        verifylist = [
            ('service', updated_service.id),
            ('registered_limit_id', self.registered_limit.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.update_registered_limit.assert_called_with(
            self.registered_limit.id,
            service_id=updated_service.id,
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            self.default_limit,
            None,
            self.registered_limit.id,
            None,
            self.resource_name,
            updated_service.id,
        )
        self.assertEqual(datalist, data)

    def test_registered_limit_set_region(self):
        updated_region = sdk_fakes.generate_fake_resource(_region.Region)
        self.identity_sdk_client.get_region.return_value = updated_region
        updated_registered_limit = sdk_fakes.generate_fake_resource(
            resource_type=_registered_limit.RegisteredLimit,
            id=self.registered_limit.id,
            description=None,
            region_id=updated_region.id,
            service_id=self.service.id,
            default_limit=self.default_limit,
            resource_name=self.resource_name,
        )
        self.identity_sdk_client.update_registered_limit.return_value = (
            updated_registered_limit
        )

        arglist = ['--region', updated_region.id, self.registered_limit.id]
        verifylist = [
            ('region', updated_region.id),
            ('registered_limit_id', self.registered_limit.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.update_registered_limit.assert_called_with(
            self.registered_limit.id,
            region_id=updated_region.id,
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            self.default_limit,
            None,
            self.registered_limit.id,
            updated_region.id,
            self.resource_name,
            self.service.id,
        )
        self.assertEqual(datalist, data)


class TestRegisteredLimitList(identity_fakes.TestIdentity):
    def setUp(self):
        super().setUp()

        self.service = sdk_fakes.generate_fake_resource(_service.Service)
        self.region = sdk_fakes.generate_fake_resource(_region.Region)

        self.description = 'default limit of foobars'
        self.default_limit = 10
        self.resource_name = 'foobars'

        self.identity_sdk_client.find_service.return_value = self.service
        self.identity_sdk_client.get_region.return_value = self.region

        self.registered_limit = sdk_fakes.generate_fake_resource(
            resource_type=_registered_limit.RegisteredLimit,
            description=None,
            region_id=None,
            service_id=self.service.id,
            default_limit=self.default_limit,
            resource_name=self.resource_name,
        )
        self.registered_limit_with_options = sdk_fakes.generate_fake_resource(
            resource_type=_registered_limit.RegisteredLimit,
            description=self.description,
            region_id=self.region.id,
            service_id=self.service.id,
            default_limit=self.default_limit,
            resource_name=self.resource_name,
        )
        self.identity_sdk_client.registered_limits.return_value = [
            self.registered_limit,
            self.registered_limit_with_options,
        ]

        self.cmd = registered_limit.ListRegisteredLimit(self.app, None)

    def test_registered_limit_list(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.registered_limits.assert_called_with()
        collist = (
            "ID",
            "Service ID",
            "Resource Name",
            "Default Limit",
            "Description",
            "Region ID",
        )
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.registered_limit.id,
                self.service.id,
                self.resource_name,
                self.default_limit,
                None,
                None,
            ),
            (
                self.registered_limit_with_options.id,
                self.service.id,
                self.resource_name,
                self.default_limit,
                self.description,
                self.region.id,
            ),
        )
        self.assertEqual(datalist, tuple(data))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest.mock import call

from openstack import exceptions as sdk_exceptions
from openstack.identity.v3 import credential as _credential
from openstack.identity.v3 import project as _project
from openstack.identity.v3 import user as _user
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.identity.v3 import credential
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes
from openstackclient.tests.unit import utils


class TestCredentialCreate(identity_fakes.TestIdentity):
    user = sdk_fakes.generate_fake_resource(_user.User)
    project = sdk_fakes.generate_fake_resource(_project.Project)
    columns = (
        'blob',
        'id',
        'project_id',
        'type',
        'user_id',
    )

    def setUp(self):
        super().setUp()

        self.credential = sdk_fakes.generate_fake_resource(
            resource_type=_credential.Credential,
            user_id=self.user.id,
            project_id=self.project.id,
            type='cert',
        )
        self.identity_sdk_client.create_credential.return_value = (
            self.credential
        )
        self.identity_sdk_client.find_user.return_value = self.user
        self.identity_sdk_client.find_project.return_value = self.project
        self.data = (
            self.credential.blob,
            self.credential.id,
            self.credential.project_id,
            self.credential.type,
            self.credential.user_id,
        )

        self.cmd = credential.CreateCredential(self.app, None)

    def test_credential_create_no_options(self):
        arglist = [
            self.credential.user_id,
            self.credential.blob,
        ]
        verifylist = [
            ('user', self.credential.user_id),
            ('data', self.credential.blob),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'user_id': self.credential.user_id,
            'type': self.credential.type,
            'blob': self.credential.blob,
            'project_id': None,
        }
        self.identity_sdk_client.create_credential.assert_called_once_with(
            **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_credential_create_with_options(self):
        arglist = [
            self.credential.user_id,
            self.credential.blob,
            '--type',
            self.credential.type,
            '--project',
            self.credential.project_id,
        ]
        verifylist = [
            ('user', self.credential.user_id),
            ('data', self.credential.blob),
            ('type', self.credential.type),
            ('project', self.credential.project_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'user_id': self.credential.user_id,
            'type': self.credential.type,
            'blob': self.credential.blob,
            'project_id': self.credential.project_id,
        }
        self.identity_sdk_client.create_credential.assert_called_once_with(
            **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestCredentialDelete(identity_fakes.TestIdentity):
    def setUp(self):
        super().setUp()

        self.identity_sdk_client.delete_credential.return_value = None

        # Get the command object to test
        self.cmd = credential.DeleteCredential(self.app, None)

    def test_credential_delete(self):
        credential = sdk_fakes.generate_fake_resource(
            _credential.Credential,
        )
        arglist = [
            credential.id,
        ]
        verifylist = [
            ('credential', [credential.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_credential.assert_called_with(
            credential.id,
        )
        self.assertIsNone(result)

    def test_credential_multi_delete(self):
        credentials = sdk_fakes.generate_fake_resources(
            _credential.Credential, count=2
        )
        arglist = []
        for c in credentials:
            arglist.append(c.id)
        verifylist = [
            ('credential', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = []
        for c in credentials:
            calls.append(call(c.id))
        self.identity_sdk_client.delete_credential.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_credential_multi_delete_with_exception(self):
        credential = sdk_fakes.generate_fake_resource(
            _credential.Credential,
        )
        arglist = [
            credential.id,
            'unexist_credential',
        ]
        verifylist = [('credential', [credential.id, 'unexist_credential'])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.identity_sdk_client.delete_credential.side_effect = [
            None,
            sdk_exceptions.NotFoundException,
        ]

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 credential failed to delete.', str(e))

        self.identity_sdk_client.delete_credential.assert_any_call(
            credential.id
        )
        self.identity_sdk_client.delete_credential.assert_any_call(
            'unexist_credential'
        )


class TestCredentialList(identity_fakes.TestIdentity):
    credential = sdk_fakes.generate_fake_resource(_credential.Credential)

    columns = ('ID', 'Type', 'User ID', 'Data', 'Project ID')
    data = (
        (
            credential.id,
            credential.type,
            credential.user_id,
            credential.blob,
            credential.project_id,
        ),
    )

    def setUp(self):
        super().setUp()

        self.user = sdk_fakes.generate_fake_resource(_user.User)
        self.identity_sdk_client.find_user.return_value = self.user

        self.identity_sdk_client.credentials.return_value = [self.credential]

        # Get the command object to test
        self.cmd = credential.ListCredential(self.app, None)

    def test_credential_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.credentials.assert_called_with()
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, tuple(data))

    def test_credential_list_with_options(self):
        arglist = [
            '--user',
            self.credential.user_id,
            '--type',
            self.credential.type,
        ]
        verifylist = [
            ('user', self.credential.user_id),
            ('type', self.credential.type),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'user_id': self.user.id,
            'type': self.credential.type,
        }
        self.identity_sdk_client.find_user.assert_called_with(
            self.credential.user_id, domain_id=None, ignore_missing=False
        )
        self.identity_sdk_client.credentials.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, tuple(data))


class TestCredentialSet(identity_fakes.TestIdentity):
    credential = sdk_fakes.generate_fake_resource(_credential.Credential)

    def setUp(self):
        super().setUp()
        self.cmd = credential.SetCredential(self.app, None)

    def test_credential_set_no_options(self):
        arglist = [
            self.credential.id,
        ]

        self.assertRaises(
            utils.ParserException, self.check_parser, self.cmd, arglist, []
        )

    def test_credential_set_missing_user(self):
        arglist = [
            '--type',
            'ec2',
            '--data',
            self.credential.blob,
            self.credential.id,
        ]

        self.assertRaises(
            utils.ParserException, self.check_parser, self.cmd, arglist, []
        )

    def test_credential_set_missing_type(self):
        arglist = [
            '--user',
            self.credential.user_id,
            '--data',
            self.credential.blob,
            self.credential.id,
        ]

        self.assertRaises(
            utils.ParserException, self.check_parser, self.cmd, arglist, []
        )

    def test_credential_set_missing_data(self):
        arglist = [
            '--user',
            self.credential.user_id,
            '--type',
            'ec2',
            self.credential.id,
        ]

        self.assertRaises(
            utils.ParserException, self.check_parser, self.cmd, arglist, []
        )

    def test_credential_set_valid(self):
        arglist = [
            '--user',
            self.credential.user_id,
            '--type',
            'ec2',
            '--data',
            self.credential.blob,
            self.credential.id,
        ]
        parsed_args = self.check_parser(self.cmd, arglist, [])

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)

    def test_credential_set_valid_with_project(self):
        arglist = [
            '--user',
            self.credential.user_id,
            '--type',
            'ec2',
            '--data',
            self.credential.blob,
            '--project',
            self.credential.project_id,
            self.credential.id,
        ]
        parsed_args = self.check_parser(self.cmd, arglist, [])

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)


class TestCredentialShow(identity_fakes.TestIdentity):
    columns = (
        'blob',
        'id',
        'project_id',
        'type',
        'user_id',
    )

    def setUp(self):
        super().setUp()

        self.credential = sdk_fakes.generate_fake_resource(
            _credential.Credential
        )
        self.identity_sdk_client.get_credential.return_value = self.credential
        self.data = (
            self.credential.blob,
            self.credential.id,
            self.credential.project_id,
            self.credential.type,
            self.credential.user_id,
        )

        self.cmd = credential.ShowCredential(self.app, None)

    def test_credential_show(self):
        arglist = [
            self.credential.id,
        ]
        verifylist = [
            ('credential', self.credential.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.get_credential.assert_called_once_with(
            self.credential.id
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest.mock import call

from openstack.block_storage.v2 import consistency_group as _consistency_group
from openstack.block_storage.v2 import (
    consistency_group_snapshot as _cg_snapshot,
)
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes
from openstackclient.volume.v2 import consistency_group_snapshot


class TestConsistencyGroupSnapshotCreate(volume_fakes.TestVolume):
    _consistency_group_snapshot = sdk_fakes.generate_fake_resource(
        _cg_snapshot.ConsistencyGroupSnapshot
    )
    consistency_group = sdk_fakes.generate_fake_resource(
        _consistency_group.ConsistencyGroup
    )

    columns = (
        'consistencygroup_id',
        'created_at',
        'description',
        'id',
        'name',
        'status',
    )
    data = (
        _consistency_group_snapshot.consistencygroup_id,
        _consistency_group_snapshot.created_at,
        _consistency_group_snapshot.description,
        _consistency_group_snapshot.id,
        _consistency_group_snapshot.name,
        _consistency_group_snapshot.status,
    )

    def setUp(self):
        super().setUp()
        self.volume_client.create_consistency_group_snapshot.return_value = (
            self._consistency_group_snapshot
        )
        self.volume_client.find_consistency_group.return_value = (
            self.consistency_group
        )

        # Get the command object to test
        self.cmd = consistency_group_snapshot.CreateConsistencyGroupSnapshot(
            self.app, None
        )

    def test_consistency_group_snapshot_create(self):
        arglist = [
            '--consistency-group',
            self.consistency_group.id,
            '--description',
            self._consistency_group_snapshot.description,
            self._consistency_group_snapshot.name,
        ]
        verifylist = [
            ('consistency_group', self.consistency_group.id),
            ('description', self._consistency_group_snapshot.description),
            ('snapshot_name', self._consistency_group_snapshot.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_consistency_group.assert_called_once_with(
            self.consistency_group.id, ignore_missing=False
        )
        self.volume_client.create_consistency_group_snapshot.assert_called_once_with(
            consistencygroup_id=self.consistency_group.id,
            name=self._consistency_group_snapshot.name,
            description=self._consistency_group_snapshot.description,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_consistency_group_snapshot_create_no_consistency_group(self):
        arglist = [
            '--description',
            self._consistency_group_snapshot.description,
            self._consistency_group_snapshot.name,
        ]
        verifylist = [
            ('description', self._consistency_group_snapshot.description),
            ('snapshot_name', self._consistency_group_snapshot.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_consistency_group.assert_called_once_with(
            self._consistency_group_snapshot.name, ignore_missing=False
        )
        self.volume_client.create_consistency_group_snapshot.assert_called_once_with(
            consistencygroup_id=self.consistency_group.id,
            name=self._consistency_group_snapshot.name,
            description=self._consistency_group_snapshot.description,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestConsistencyGroupSnapshotDelete(volume_fakes.TestVolume):
    consistency_group_snapshots = [
        sdk_fakes.generate_fake_resource(
            _cg_snapshot.ConsistencyGroupSnapshot
        ),
        sdk_fakes.generate_fake_resource(
            _cg_snapshot.ConsistencyGroupSnapshot
        ),
    ]

    def setUp(self):
        super().setUp()

        self.volume_client.find_consistency_group_snapshot.side_effect = (
            self.consistency_group_snapshots
        )
        self.volume_client.delete_consistency_group_snapshot.return_value = (
            None
        )

        # Get the command object to mock
        self.cmd = consistency_group_snapshot.DeleteConsistencyGroupSnapshot(
            self.app, None
        )

    def test_consistency_group_snapshot_delete(self):
        arglist = [self.consistency_group_snapshots[0].id]
        verifylist = [
            (
                "consistency_group_snapshot",
                [self.consistency_group_snapshots[0].id],
            )
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.find_consistency_group_snapshot.assert_called_once_with(
            self.consistency_group_snapshots[0].id, ignore_missing=False
        )
        self.volume_client.delete_consistency_group_snapshot.assert_called_once_with(
            self.consistency_group_snapshots[0]
        )
        self.assertIsNone(result)

    def test_multiple_consistency_group_snapshots_delete(self):
        arglist = []
        for c in self.consistency_group_snapshots:
            arglist.append(c.id)
        verifylist = [
            ('consistency_group_snapshot', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        find_calls = []
        delete_calls = []
        for c in self.consistency_group_snapshots:
            find_calls.append(call(c.id, ignore_missing=False))
            delete_calls.append(call(c))
        self.volume_client.find_consistency_group_snapshot.assert_has_calls(
            find_calls
        )
        self.volume_client.delete_consistency_group_snapshot.assert_has_calls(
            delete_calls
        )
        self.assertIsNone(result)

    def test_delete_with_exception(self):
        arglist = ['missing-snapshot']
        verifylist = [('consistency_group_snapshot', ['missing-snapshot'])]

        self.volume_client.find_consistency_group_snapshot.side_effect = (
            exceptions.CommandError
        )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestConsistencyGroupSnapshotList(volume_fakes.TestVolume):
    consistency_group_snapshots = [
        sdk_fakes.generate_fake_resource(
            _cg_snapshot.ConsistencyGroupSnapshot, status='available'
        ),
        sdk_fakes.generate_fake_resource(
            _cg_snapshot.ConsistencyGroupSnapshot, status='available'
        ),
    ]
    consistency_group = sdk_fakes.generate_fake_resource(
        _consistency_group.ConsistencyGroup
    )

    column_headers = [
        'ID',
        'Status',
        'Name',
    ]
    column_headers_long = [
        'ID',
        'Status',
        'ConsistencyGroup ID',
        'Name',
        'Description',
        'Created At',
    ]
    data = []
    for c in consistency_group_snapshots:
        data.append(
            (
                c.id,
                c.status,
                c.name,
            )
        )
    data_long = []
    for c in consistency_group_snapshots:
        data_long.append(
            (
                c.id,
                c.status,
                c.consistencygroup_id,
                c.name,
                c.description,
                c.created_at,
            )
        )

    def setUp(self):
        super().setUp()

        self.volume_client.consistency_group_snapshots.return_value = (
            self.consistency_group_snapshots
        )
        self.volume_client.find_consistency_group.return_value = (
            self.consistency_group
        )
        # Get the command to test
        self.cmd = consistency_group_snapshot.ListConsistencyGroupSnapshot(
            self.app, None
        )

    def test_consistency_group_snapshot_list_without_options(self):
        arglist = []
        verifylist = [
            ("all_projects", False),
            ("long", False),
            ("status", None),
            ("consistency_group", None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.consistency_group_snapshots.assert_called_once_with(
            all_tenants=False,
            status=None,
            consistencygroup_id=None,
        )
        self.assertEqual(self.column_headers, columns)
        self.assertEqual(self.data, list(data))

    def test_consistency_group_snapshot_list_with_long(self):
        arglist = [
            "--long",
        ]
        verifylist = [
            ("all_projects", False),
            ("long", True),
            ("status", None),
            ("consistency_group", None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.consistency_group_snapshots.assert_called_once_with(
            all_tenants=False,
            status=None,
            consistencygroup_id=None,
        )
        self.assertEqual(self.column_headers_long, columns)
        self.assertEqual(self.data_long, list(data))

    def test_consistency_group_snapshot_list_with_options(self):
        arglist = [
            "--all-project",
            "--status",
            self.consistency_group_snapshots[0].status,
            "--consistency-group",
            self.consistency_group.id,
        ]
        verifylist = [
            ("all_projects", True),
            ("long", False),
            ("status", self.consistency_group_snapshots[0].status),
            ("consistency_group", self.consistency_group.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_consistency_group.assert_called_once_with(
            self.consistency_group.id, ignore_missing=False
        )
        self.volume_client.consistency_group_snapshots.assert_called_once_with(
            all_tenants=True,
            status=self.consistency_group_snapshots[0].status,
            consistencygroup_id=self.consistency_group.id,
        )
        self.assertEqual(self.column_headers, columns)
        self.assertEqual(self.data, list(data))


class TestConsistencyGroupSnapshotShow(volume_fakes.TestVolume):
    _consistency_group_snapshot = sdk_fakes.generate_fake_resource(
        _cg_snapshot.ConsistencyGroupSnapshot
    )

    columns = (
        'consistencygroup_id',
        'created_at',
        'description',
        'id',
        'name',
        'status',
    )
    data = (
        _consistency_group_snapshot.consistencygroup_id,
        _consistency_group_snapshot.created_at,
        _consistency_group_snapshot.description,
        _consistency_group_snapshot.id,
        _consistency_group_snapshot.name,
        _consistency_group_snapshot.status,
    )

    def setUp(self):
        super().setUp()

        self.volume_client.find_consistency_group_snapshot.return_value = (
            self._consistency_group_snapshot
        )
        self.cmd = consistency_group_snapshot.ShowConsistencyGroupSnapshot(
            self.app, None
        )

    def test_consistency_group_snapshot_show(self):
        arglist = [self._consistency_group_snapshot.id]
        verifylist = [
            ("consistency_group_snapshot", self._consistency_group_snapshot.id)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.find_consistency_group_snapshot.assert_called_once_with(
            self._consistency_group_snapshot.id, ignore_missing=False
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Storage backend action implementations"""

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from openstack import utils as sdk_utils
from osc_lib.cli import format_columns
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


class ShowCapability(command.Lister):
    _description = _("Show capability command")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "host",
            metavar="<host>",
            help=_("List capabilities of specified host (host@backend-name)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )

        columns = [
            'Title',
            'Key',
            'Type',
            'Description',
        ]

        data = volume_client.get_capabilities(parsed_args.host)

        # The get capabilities API is... interesting. We only want the names of
        # the capabilities that can set for a backend through extra specs, so
        # we need to extract out that part of the mess that is returned.
        print_data = []
        keys = data.properties
        for key in keys:
            # Stuff the key into the details to make it easier to output
            capability_data = data.properties[key]
            capability_data['key'] = key
            print_data.append(capability_data)

        return (
            columns,
            (
                utils.get_dict_properties(
                    s,
                    columns,
                )
                for s in print_data
            ),
        )


class ListPool(command.Lister):
    _description = _("List pool command")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--long",
            action="store_true",
            default=False,
            help=_("Show detailed information about pools."),
        )
        # TODO(smcginnis): Starting with Cinder microversion 3.33, user is also
        # able to pass in --filters with a <key>=<value> pair to filter on.
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )

        if parsed_args.long:
            columns = [
                'name',
                'capabilities',
            ]

            headers = [
                'Name',
                'Capabilities',
            ]
        else:
            columns = [
                'Name',
            ]
            headers = columns

        data = volume_client.backend_pools(detailed=parsed_args.long)
        formatters = {'capabilities': format_columns.DictColumn}
        return (
            headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters=formatters,
                )
                for s in data
            ),
        )



# Copyright (c) 2016, Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from typing import ClassVar
import unittest
import uuid

from openstackclient.tests.functional.network.v2 import common


class TestMeterRule(common.NetworkTests):
    """Functional tests for meter rule"""

    METER_ID: ClassVar[str]
    METER_NAME: ClassVar[str]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        if not cls.is_extension_enabled("metering"):
            raise unittest.SkipTest("No metering extension present")

        if cls.haz_network:
            cls.METER_NAME = uuid.uuid4().hex

            json_output = cls.openstack(
                'network meter create ' + cls.METER_NAME,
                parse_output=True,
            )
            cls.METER_ID = json_output.get('id')

    @classmethod
    def tearDownClass(cls):
        try:
            if cls.haz_network:
                raw_output = cls.openstack(
                    'network meter delete ' + cls.METER_ID
                )
                cls.assertOutput('', raw_output)
        finally:
            super().tearDownClass()

    def test_meter_rule_delete(self):
        """test create, delete"""
        json_output = self.openstack(
            'network meter rule create '
            + '--remote-ip-prefix 10.0.0.0/8 '
            + self.METER_ID,
            parse_output=True,
        )
        rule_id = json_output.get('id')
        re_ip = json_output.get('remote_ip_prefix')

        self.addCleanup(self.openstack, 'network meter rule delete ' + rule_id)
        self.assertIsNotNone(re_ip)
        self.assertIsNotNone(rule_id)
        self.assertEqual('10.0.0.0/8', re_ip)

    def test_meter_rule_list(self):
        """Test create, list, delete"""
        json_output = self.openstack(
            'network meter rule create '
            + '--remote-ip-prefix 10.0.0.0/8 '
            + self.METER_ID,
            parse_output=True,
        )
        rule_id_1 = json_output.get('id')
        self.addCleanup(
            self.openstack, 'network meter rule delete ' + rule_id_1
        )
        self.assertEqual('10.0.0.0/8', json_output.get('remote_ip_prefix'))

        json_output_1 = self.openstack(
            'network meter rule create '
            + '--remote-ip-prefix 11.0.0.0/8 '
            + self.METER_ID,
            parse_output=True,
        )
        rule_id_2 = json_output_1.get('id')
        self.addCleanup(
            self.openstack, 'network meter rule delete ' + rule_id_2
        )
        self.assertEqual('11.0.0.0/8', json_output_1.get('remote_ip_prefix'))

        json_output = self.openstack(
            'network meter rule list',
            parse_output=True,
        )
        rule_id_list = [item.get('ID') for item in json_output]
        ip_prefix_list = [item.get('Remote IP Prefix') for item in json_output]
        self.assertIn(rule_id_1, rule_id_list)
        self.assertIn(rule_id_2, rule_id_list)
        self.assertIn('10.0.0.0/8', ip_prefix_list)
        self.assertIn('11.0.0.0/8', ip_prefix_list)

    def test_meter_rule_show(self):
        """Test create, show, delete"""
        json_output = self.openstack(
            'network meter rule create '
            + '--remote-ip-prefix 10.0.0.0/8 '
            + '--egress '
            + self.METER_ID,
            parse_output=True,
        )
        rule_id = json_output.get('id')

        self.assertEqual('egress', json_output.get('direction'))

        json_output = self.openstack(
            'network meter rule show ' + rule_id,
            parse_output=True,
        )
        self.assertEqual('10.0.0.0/8', json_output.get('remote_ip_prefix'))
        self.assertIsNotNone(rule_id)

        self.addCleanup(self.openstack, 'network meter rule delete ' + rule_id)



#   Copyright 2023 Red Hat
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Image V2 Action Implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack.image.v2 import metadef_object as _metadef_object
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


def _format_object(
    md_object: _metadef_object.MetadefObject,
) -> tuple[tuple[str, ...], Iterable[Any]]:
    fields_to_show = (
        'created_at',
        'description',
        'name',
        'namespace_name',
        'properties',
        'required',
        'updated_at',
    )

    return (
        fields_to_show,
        utils.get_item_properties(md_object, fields_to_show),
    )


class CreateMetadefObjects(command.ShowOne):
    _description = _("Create a metadef object")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--namespace",
            metavar="<namespace>",
            help=_("Metadef namespace to create the object in (name)"),
        )
        parser.add_argument(
            "name",
            metavar='<metadef-object-name>',
            help=_('New metadef object name'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        image_client = self.app.client_manager.image

        namespace = image_client.get_metadef_namespace(
            parsed_args.namespace,
        )
        data = image_client.create_metadef_object(
            namespace=namespace.namespace,
            name=parsed_args.name,
        )

        fields, value = _format_object(data)

        return fields, value


class ShowMetadefObjects(command.ShowOne):
    _description = _("Show a particular metadef object")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "namespace",
            metavar="<namespace>",
            help=_("Metadef namespace of the object (name)"),
        )
        parser.add_argument(
            "object",
            metavar="<object>",
            help=_("Metadef object to show"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        image_client = self.app.client_manager.image

        namespace = parsed_args.namespace
        object = parsed_args.object

        data = image_client.get_metadef_object(object, namespace)

        fields, value = _format_object(data)

        return fields, value


class DeleteMetadefObject(command.Command):
    _description = _("Delete metadata definitions object(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "namespace",
            metavar="<namespace>",
            help=_("Metadef namespace of the object (name)"),
        )
        parser.add_argument(
            "objects",
            metavar="<object>",
            nargs="*",
            help=_(
                "Metadef object(s) to delete (name) "
                "(omit this argument to delete all objects in the namespace)"
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        image_client = self.app.client_manager.image

        namespace = parsed_args.namespace

        if not parsed_args.objects:
            image_client.delete_all_metadef_objects(namespace)
            return

        result = 0
        for obj in parsed_args.objects:
            try:
                object = image_client.get_metadef_object(obj, namespace)
                image_client.delete_metadef_object(object, namespace)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete object with name or "
                        "ID '%(object)s': %(e)s"
                    ),
                    {'object': obj, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.namespace)
            msg = _("%(result)s of %(total)s object failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListMetadefObjects(command.Lister):
    _description = _("List metadef objects inside a specific namespace.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "namespace",
            metavar="<namespace>",
            help=_("Namespace (name) for the namespace"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        image_client = self.app.client_manager.image

        namespace = parsed_args.namespace
        columns = ['name', 'description']

        md_objects = list(image_client.metadef_objects(namespace))
        column_headers = columns
        return (
            column_headers,
            (
                utils.get_item_properties(
                    md_object,
                    columns,
                )
                for md_object in md_objects
            ),
        )


class SetMetadefObject(command.Command):
    _description = _("Update a metadef object")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "namespace",
            metavar="<namespace>",
            help=_("Metadef namespace name"),
        )
        parser.add_argument(
            "object",
            metavar="<object>",
            help=_('Metadef object to be updated'),
        )
        parser.add_argument(
            "--name",
            help=_("New name of the object"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        image_client = self.app.client_manager.image

        object = image_client.get_metadef_object(
            parsed_args.object, parsed_args.namespace
        )
        kwargs = {}
        if parsed_args.name:
            kwargs['name'] = parsed_args.name

        image_client.update_metadef_object(
            object, parsed_args.namespace, **kwargs
        )


class ShowMetadefObjectProperty(command.ShowOne):
    _description = _(
        "Describe a specific metadata definitions property inside an object."
    )

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "namespace",
            metavar="<namespace_name>",
            help=_("Namespace (name) for the namespace"),
        )
        parser.add_argument(
            "object",
            metavar="<object_name>",
            help=_("Name of an object."),
        )
        parser.add_argument(
            "property",
            help=_("Name of the property."),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        image_client = self.app.client_manager.image

        namespace_name = parsed_args.namespace
        object_name = parsed_args.object

        obj = image_client.get_metadef_object(object_name, namespace_name)
        try:
            prop = obj['properties'][parsed_args.property]
            prop['name'] = parsed_args.property

        except KeyError:
            msg = _(
                'Property %(property)s not found in object %(object)s.'
            ) % {
                'property': parsed_args.property,
                'object': parsed_args.object,
            }
            raise exceptions.CommandError(msg)

        col_headers, col_data = zip(*sorted(prop.items()))
        return col_headers, col_data



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from manilaclient import api_versions
from osc_lib import exceptions
from osc_lib import utils as osc_utils

from openstackclient import command
from openstackclient.i18n import _


class SetShareService(command.Command):
    """Enable/disable share service (Admin only)."""

    _description = _("Enable/Disable share service (Admin only).")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'host',
            metavar='<host>',
            help=_("Host name as 'example_host@example_backend'."),
        )
        parser.add_argument(
            'binary',
            metavar='<binary>',
            help=_(
                "Service binary, could be 'manila-share', "
                "'manila-scheduler' or 'manila-data'"
            ),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            action='store_true',
            help=_('Enable share service'),
        )
        enable_group.add_argument(
            '--disable',
            action='store_true',
            help=_('Disable share service'),
        )
        parser.add_argument(
            "--disable-reason",
            metavar="<reason>",
            help=_(
                "Reason for disabling the service "
                "(should be used with --disable option)"
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        if parsed_args.disable_reason and not parsed_args.disable:
            msg = _(
                "Cannot specify option --disable-reason without "
                "--disable specified."
            )
            raise exceptions.CommandError(msg)

        share_client = self.app.client_manager.share

        if parsed_args.enable:
            try:
                share_client.services.enable(
                    parsed_args.host, parsed_args.binary
                )
            except Exception as e:
                msg = _("Failed to enable service: %(e)s")
                raise exceptions.CommandError(msg % {'e': e})

        if parsed_args.disable:
            if parsed_args.disable_reason:
                if share_client.api_version < api_versions.APIVersion("2.83"):
                    raise exceptions.CommandError(
                        "Service disable reason can be specified only with "
                        "manila API version >= 2.83"
                    )
            try:
                if parsed_args.disable_reason:
                    share_client.services.disable(
                        parsed_args.host,
                        parsed_args.binary,
                        disable_reason=parsed_args.disable_reason,
                    )
                else:
                    share_client.services.disable(
                        parsed_args.host, parsed_args.binary
                    )
            except Exception as e:
                msg = _("Failed to disable service: %(e)s")
                raise exceptions.CommandError(msg % {'e': e})


class ListShareService(command.Lister):
    """List share services (Admin only)."""

    _description = _("List share services (Admin only).")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--host",
            metavar="<host>",
            default=None,
            help=_("Filter services by name of the host."),
        )
        parser.add_argument(
            "--binary",
            metavar="<binary>",
            default=None,
            help=_("Filter services by the name of the service."),
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            default=None,
            help=_("Filter results by status."),
        )
        parser.add_argument(
            "--state",
            metavar="<state>",
            default=None,
            choices=['up', 'down'],
            help=_("Filter results by state."),
        )
        parser.add_argument(
            "--zone",
            metavar="<zone>",
            default=None,
            help=_("Filter services by their availability zone."),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        share_client = self.app.client_manager.share

        search_opts = {
            'host': parsed_args.host,
            'binary': parsed_args.binary,
            'status': parsed_args.status,
            'state': parsed_args.state,
            'zone': parsed_args.zone,
        }

        services = share_client.services.list(search_opts=search_opts)

        columns = [
            'ID',
            'Binary',
            'Host',
            'Zone',
            'Status',
            'State',
            'Updated At',
        ]
        if share_client.api_version >= api_versions.APIVersion("2.83"):
            columns.append('Disabled Reason')
        if share_client.api_version >= api_versions.APIVersion("2.86"):
            columns.append('Ensuring')

        data = (
            osc_utils.get_dict_properties(service._info, columns)
            for service in services
        )

        return (columns, data)


class EnsureShareService(command.Command):
    """Run ensure shares in a back end (Admin only)."""

    _description = _("Run ensure shares in a back end (Admin only).")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'host',
            metavar='<host>',
            help=_(
                "Host to run ensure shares. 'example_host@example_backend'."
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share

        if share_client.api_version < api_versions.APIVersion("2.86"):
            raise exceptions.CommandError(
                "Ensure shares API is only available in "
                "manila API version >= 2.86"
            )

        try:
            share_client.services.ensure_shares(parsed_args.host)
        except Exception as e:
            msg = _("Failed to ensure shares: %(e)s")
            raise exceptions.CommandError(msg % {'e': e})



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock
import uuid

from manilaclient import api_versions
from osc_lib import exceptions
from osc_lib import exceptions as osc_exceptions
from osc_lib import utils as oscutils

from openstackclient.share import utils
from openstackclient.share.v2 import share_groups
from openstackclient.tests.unit.share.v2 import fakes as share_fakes
from openstackclient.tests.unit import utils as test_utils


class TestShareGroup(share_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.groups_mock = self.share_client.share_groups
        self.groups_mock.reset_mock()

        self.share_types_mock = self.share_client.share_types
        self.share_types_mock.reset_mock()

        self.set_share_api_version(api_versions.MAX_VERSION)


class TestShareGroupCreate(TestShareGroup):
    def setUp(self):
        super().setUp()

        self.share_group = share_fakes.FakeShareGroup.create_one_share_group()
        self.formatted_result = (
            share_fakes.FakeShareGroup.create_one_share_group(
                attrs={
                    "id": self.share_group.id,
                    'created_at': self.share_group.created_at,
                    "project_id": self.share_group.project_id,
                    'share_group_type_id': (
                        self.share_group.share_group_type_id
                    ),
                    'share_types': '\n'.join(self.share_group.share_types),
                }
            )
        )

        self.groups_mock.create.return_value = self.share_group
        self.groups_mock.get.return_value = self.share_group

        self.cmd = share_groups.CreateShareGroup(self.app, None)

        self.data = tuple(self.formatted_result._info.values())
        self.columns = tuple(self.share_group._info.keys())

    def test_share_group_create_no_args(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.groups_mock.create.assert_called_with(
            name=None,
            description=None,
            share_types=[],
            share_group_type=None,
            share_network=None,
            source_share_group_snapshot=None,
            availability_zone=None,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_group_create_with_options(self):
        arglist = [
            '--name',
            self.share_group.name,
            '--description',
            self.share_group.description,
        ]
        verifylist = [
            ('name', self.share_group.name),
            ('description', self.share_group.description),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.groups_mock.create.assert_called_with(
            name=self.share_group.name,
            description=self.share_group.description,
            share_types=[],
            share_group_type=None,
            share_network=None,
            source_share_group_snapshot=None,
            availability_zone=None,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_group_create_az(self):
        arglist = ['--availability-zone', self.share_group.availability_zone]
        verifylist = [
            ('availability_zone', self.share_group.availability_zone)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.groups_mock.create.assert_called_with(
            name=None,
            description=None,
            share_types=[],
            share_group_type=None,
            share_network=None,
            source_share_group_snapshot=None,
            availability_zone=self.share_group.availability_zone,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_group_create_share_types(self):
        share_types = share_fakes.FakeShareType.create_share_types(count=2)
        self.share_types_mock.get = share_fakes.FakeShareType.get_share_types(
            share_types
        )
        arglist = ['--share-types', share_types[0].id, share_types[1].id]
        verifylist = [('share_types', [share_types[0].id, share_types[1].id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.groups_mock.create.assert_called_with(
            name=None,
            description=None,
            share_types=share_types,
            share_group_type=None,
            share_network=None,
            source_share_group_snapshot=None,
            availability_zone=None,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_group_create_wait(self):
        arglist = ['--wait']
        verifylist = [('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.groups_mock.create.assert_called_with(
            name=None,
            description=None,
            share_types=[],
            share_group_type=None,
            share_network=None,
            source_share_group_snapshot=None,
            availability_zone=None,
        )

        self.groups_mock.get.assert_called_with(self.share_group.id)
        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    # TODO(archanaserver): Add test cases for share-group-type,
    # share-network and source-share-group-snapshot when the
    # options have OSC support.


class TestShareGroupDelete(TestShareGroup):
    def setUp(self):
        super().setUp()

        self.share_group = share_fakes.FakeShareGroup.create_one_share_group()
        self.groups_mock.get.return_value = self.share_group

        self.cmd = share_groups.DeleteShareGroup(self.app, None)

    def test_share_group_delete(self):
        arglist = [self.share_group.id]
        verifylist = [('share_group', [self.share_group.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.groups_mock.delete.assert_called_with(
            self.share_group, force=False
        )
        self.assertIsNone(result)

    def test_share_group_delete_force(self):
        arglist = [self.share_group.id, '--force']
        verifylist = [('share_group', [self.share_group.id]), ('force', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.groups_mock.delete.assert_called_with(
            self.share_group, force=True
        )
        self.assertIsNone(result)

    def test_share_group_delete_multiple(self):
        share_groups = share_fakes.FakeShareGroup.create_share_groups(count=2)
        arglist = [share_groups[0].id, share_groups[1].id]
        verifylist = [
            ('share_group', [share_groups[0].id, share_groups[1].id])
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertEqual(self.groups_mock.delete.call_count, len(share_groups))
        self.assertIsNone(result)

    def test_share_group_delete_exception(self):
        arglist = [self.share_group.id]
        verifylist = [('share_group', [self.share_group.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.groups_mock.delete.side_effect = exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_group_delete_wait(self):
        arglist = [self.share_group.id, '--wait']
        verifylist = [('share_group', [self.share_group.id]), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_delete', return_value=True):
            result = self.cmd.take_action(parsed_args)

            self.groups_mock.delete.assert_called_with(
                self.share_group, force=False
            )
            self.groups_mock.get.assert_called_with(self.share_group.id)
            self.assertIsNone(result)

    def test_share_group_delete_wait_exception(self):
        arglist = [self.share_group.id, '--wait']
        verifylist = [('share_group', [self.share_group.id]), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_delete', return_value=False):
            self.assertRaises(
                exceptions.CommandError, self.cmd.take_action, parsed_args
            )


class TestShareGroupShow(TestShareGroup):
    def setUp(self):
        super().setUp()

        self.share_group = share_fakes.FakeShareGroup.create_one_share_group()
        self.formatted_result = (
            share_fakes.FakeShareGroup.create_one_share_group(
                attrs={
                    "id": self.share_group.id,
                    'created_at': self.share_group.created_at,
                    "project_id": self.share_group.project_id,
                    'share_group_type_id': (
                        self.share_group.share_group_type_id
                    ),
                    'share_types': '\n'.join(self.share_group.share_types),
                }
            )
        )
        self.groups_mock.get.return_value = self.share_group

        self.data = tuple(self.formatted_result._info.values())
        self.columns = tuple(self.share_group._info.keys())

        self.cmd = share_groups.ShowShareGroup(self.app, None)

    def test_share_group_show(self):
        arglist = [self.share_group.id]
        verifylist = [('share_group', self.share_group.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.groups_mock.get.assert_called_with(self.share_group.id)

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestShareGroupSet(TestShareGroup):
    def setUp(self):
        super().setUp()

        self.share_group = share_fakes.FakeShareGroup.create_one_share_group()
        self.share_group = share_fakes.FakeShare.create_one_share(
            methods={"reset_state": None}
        )
        self.groups_mock.get.return_value = self.share_group

        self.cmd = share_groups.SetShareGroup(self.app, None)

    def test_set_share_group_name(self):
        new_name = uuid.uuid4().hex
        arglist = [
            '--name',
            new_name,
            self.share_group.id,
        ]
        verifylist = [('name', new_name), ('share_group', self.share_group.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.groups_mock.update.assert_called_with(
            self.share_group.id, name=parsed_args.name
        )

    def test_set_share_group_description(self):
        new_description = uuid.uuid4().hex
        arglist = [
            '--description',
            new_description,
            self.share_group.id,
        ]
        verifylist = [
            ('description', new_description),
            ('share_group', self.share_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.groups_mock.update.assert_called_with(
            self.share_group.id, description=parsed_args.description
        )

    def test_share_group_set_status(self):
        new_status = 'available'
        arglist = [self.share_group.id, '--status', new_status]
        verifylist = [
            ('share_group', self.share_group.id),
            ('status', new_status),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.share_group.reset_state.assert_called_with(new_status)
        self.assertIsNone(result)

    def test_share_group_set_status_exception(self):
        new_status = 'available'
        arglist = [self.share_group.id, '--status', new_status]
        verifylist = [
            ('share_group', self.share_group.id),
            ('status', new_status),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.share_group.reset_state.side_effect = Exception()
        self.assertRaises(
            osc_exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareGroupUnset(TestShareGroup):
    def setUp(self):
        super().setUp()

        self.share_group = share_fakes.FakeShareGroup.create_one_share_group()
        self.groups_mock.get.return_value = self.share_group

        self.cmd = share_groups.UnsetShareGroup(self.app, None)

    def test_unset_share_group_name(self):
        arglist = [self.share_group.id, '--name']
        verifylist = [('share_group', self.share_group.id), ('name', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.groups_mock.update.assert_called_with(self.share_group, name=None)
        self.assertIsNone(result)

    def test_unset_share_group_description(self):
        arglist = [self.share_group.id, '--description']
        verifylist = [
            ('share_group', self.share_group.id),
            ('description', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.groups_mock.update.assert_called_with(
            self.share_group, description=None
        )
        self.assertIsNone(result)

    def test_unset_share_group_name_exception(self):
        arglist = [
            self.share_group.id,
            '--name',
        ]
        verifylist = [
            ('share_group', self.share_group.id),
            ('name', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.groups_mock.update.side_effect = Exception()

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareGroupList(TestShareGroup):
    columns = ['id', 'name', 'status', 'description']

    column_headers = utils.format_column_headers(columns)

    def setUp(self):
        super().setUp()

        self.new_share_group = (
            share_fakes.FakeShareGroup.create_one_share_group()
        )
        self.groups_mock.list.return_value = [self.new_share_group]

        self.share_group = share_fakes.FakeShareGroup.create_one_share_group()
        self.groups_mock.get.return_value = self.share_group

        self.share_groups_list = (
            share_fakes.FakeShareGroup.create_share_groups(count=2)
        )
        self.groups_mock.list.return_value = self.share_groups_list

        self.values = (
            oscutils.get_dict_properties(s._info, self.columns)
            for s in self.share_groups_list
        )

        self.cmd = share_groups.ListShareGroup(self.app, None)

    def test_share_group_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.groups_mock.list.assert_called_with(
            search_opts={
                'all_tenants': False,
                'name': None,
                'status': None,
                'share_server_id': None,
                'share_group_type': None,
                'snapshot': None,
                'host': None,
                'share_network': None,
                'project_id': None,
                'limit': None,
                'offset': None,
                'name~': None,
                'description~': None,
                'description': None,
            }
        )

        self.assertEqual(self.column_headers, columns)
        self.assertEqual(list(self.values), list(data))

    def test_share_group_list_pre_v236(self):
        self.set_share_api_version('2.35')

        arglist = ['--description', 'Description']
        verifylist = [('description', 'Description')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_group_lists_all_projects(self):
        all_tenants_list = self.column_headers.copy()
        all_tenants_list.append('Project ID')
        list_values = (
            oscutils.get_dict_properties(s._info, all_tenants_list)
            for s in self.share_groups_list
        )

        arglist = ['--all-projects']

        verifylist = [('all_projects', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.groups_mock.list.assert_called_with(
            search_opts={
                'all_tenants': True,
                'name': None,
                'status': None,
                'share_server_id': None,
                'share_group_type': None,
                'snapshot': None,
                'host': None,
                'share_network': None,
                'project_id': None,
                'limit': None,
                'offset': None,
                'name~': None,
                'description~': None,
                'description': None,
            }
        )

        self.assertEqual(all_tenants_list, columns)
        self.assertEqual(list(list_values), list(data))

    def test_share_group_list_name(self):
        arglist = ['--name', self.new_share_group.name]
        verifylist = [('name', self.new_share_group.name)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        search_opts = {
            'all_tenants': False,
            'name': None,
            'status': None,
            'share_server_id': None,
            'share_group_type': None,
            'snapshot': None,
            'host': None,
            'share_network': None,
            'project_id': None,
            'limit': None,
            'offset': None,
            'name~': None,
            'description~': None,
            'description': None,
        }

        search_opts['name'] = self.new_share_group.name

        self.groups_mock.list.assert_called_once_with(
            search_opts=search_opts,
        )

        self.assertEqual(self.column_headers, columns)
        self.assertEqual(list(self.values), list(data))

    def test_share_group_list_description(self):
        arglist = ['--description', self.new_share_group.description]
        verifylist = [('description', self.new_share_group.description)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        search_opts = {
            'all_tenants': False,
            'name': None,
            'status': None,
            'share_server_id': None,
            'share_group_type': None,
            'snapshot': None,
            'host': None,
            'share_network': None,
            'project_id': None,
            'limit': None,
            'offset': None,
            'name~': None,
            'description~': None,
            'description': None,
        }

        search_opts['description'] = self.new_share_group.description

        self.groups_mock.list.assert_called_once_with(
            search_opts=search_opts,
        )

        self.assertEqual(self.column_headers, columns)
        self.assertEqual(list(self.values), list(data))

    def test_share_group_list_status(self):
        arglist = [
            '--status',
            self.new_share_group.status,
        ]
        verifylist = [
            ('status', self.new_share_group.status),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        search_opts = {
            'all_tenants': False,
            'name': None,
            'status': None,
            'share_server_id': None,
            'share_group_type': None,
            'snapshot': None,
            'host': None,
            'share_network': None,
            'project_id': None,
            'limit': None,
            'offset': None,
            'name~': None,
            'description~': None,
            'description': None,
        }

        search_opts['status'] = self.new_share_group.status

        self.groups_mock.list.assert_called_once_with(
            search_opts=search_opts,
        )

        self.assertEqual(self.column_headers, columns)
        self.assertEqual(list(self.values), list(data))

    def test_share_group_list_marker_and_limit(self):
        arglist = [
            "--marker",
            self.new_share_group.id,
            "--limit",
            "2",
        ]
        verifylist = [
            ('marker', self.new_share_group.id),
            ('limit', 2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        search_opts = {
            'all_tenants': False,
            'name': None,
            'status': None,
            'share_server_id': None,
            'share_group_type': None,
            'snapshot': None,
            'host': None,
            'share_network': None,
            'project_id': None,
            'limit': 2,
            'offset': self.new_share_group.id,
            'name~': None,
            'description~': None,
            'description': None,
        }

        self.groups_mock.list.assert_called_once_with(
            search_opts=search_opts,
        )

        self.assertEqual(self.column_headers, columns)
        self.assertEqual(list(self.values), list(data))

    def test_share_group_list_negative_limit(self):
        arglist = [
            "--limit",
            "-2",
        ]
        verifylist = [
            ("limit", -2),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    # TODO(archanaserver): Add test cases for share-server-id,
    # share-group-type, snapshot, share-network and source-
    # share-group-share_group when the options have OSC support.



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
import logging
from typing import TYPE_CHECKING, Any

from osc_lib import exceptions
from osc_lib import utils

from openstackclient.i18n import _

if TYPE_CHECKING:
    from manilaclient import client

LOG = logging.getLogger(__name__)

# global variables used when building the shell
DEFAULT_API_VERSION = '2'
API_VERSION_OPTION = 'os_share_api_version'
API_NAME = 'share'

# Save the microversion if in use
_share_api_version = None


def make_client(instance: Any) -> 'client.Client':
    """Returns a manilaclient.client.Client instance."""
    # Defer client imports until we actually need them
    from manilaclient import client

    if _share_api_version is not None:
        version = _share_api_version
    else:
        from manilaclient import api_versions

        check_version = instance._api_version[API_NAME]
        version = api_versions.get_api_version(check_version)

    LOG.debug('Instantiating Shared File System client: %s', client.Client)
    LOG.debug('Shared File System API version: %s', version)

    return client.Client(
        version,
        session=instance.session,
        endpoint_type=instance.interface,
        region_name=instance.region_name,
        auth=instance.auth,
        cacert=instance.cacert,
        cert=instance.cert,
        insecure=not instance.verify,
    )


def build_option_parser(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    """Hook to add global options"""
    parser.add_argument(
        '--os-share-api-version',
        metavar='<share-api-version>',
        default=utils.env('OS_SHARE_API_VERSION'),
        help=_(
            "Shared File System API version, default=%s "
            "(Env: OS_SHARE_API_VERSION)"
        )
        % DEFAULT_API_VERSION,
    )
    return parser


def check_api_version(check_version: str) -> bool:
    # Defer client imports until we actually need them
    from manilaclient import api_versions

    global _share_api_version

    # TODO(stephenfin): Other clients (novaclient, cinderclient, ...) do this
    # normalization for us
    if check_version.isdigit():
        check_version = f"{check_version}.0"

    _share_api_version = api_versions.get_api_version(check_version)

    # Bypass X.latest format microversion
    if not _share_api_version.is_latest():
        if _share_api_version > api_versions.APIVersion("2.0"):
            if not _share_api_version.matches(
                api_versions.MIN_VERSION,
                api_versions.MAX_VERSION,
            ):
                msg = _("versions supported by client: %(min)s - %(max)s") % {
                    "min": api_versions.MIN_VERSION,
                    "max": api_versions.MAX_VERSION,
                }
                raise exceptions.CommandError(msg)

            return True

    return False



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

# from openstackclient.tests.functional import base
from openstackclient.tests.functional.object.v1 import common


class ContainerTests(common.ObjectStoreTests):
    """Functional tests for Object Store container commands"""

    NAME = uuid.uuid4().hex

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        if cls.haz_object_store:
            opts = cls.get_opts(['container'])
            raw_output = cls.openstack('container create ' + cls.NAME + opts)
            cls.assertOutput(cls.NAME + '\n', raw_output)

    @classmethod
    def tearDownClass(cls):
        try:
            if cls.haz_object_store:
                raw_output = cls.openstack('container delete ' + cls.NAME)
                cls.assertOutput('', raw_output)
        finally:
            super().tearDownClass()

    def setUp(self):
        super().setUp()
        # Skip tests if no object-store is present
        if not self.haz_object_store:
            self.skipTest("No object-store service present")

    def test_container_list(self):
        opts = self.get_opts(['Name'])
        raw_output = self.openstack('container list' + opts)
        self.assertIn(self.NAME, raw_output)

    def test_container_show(self):
        opts = self.get_opts(['container'])
        raw_output = self.openstack('container show ' + self.NAME + opts)
        self.assertEqual(self.NAME + "\n", raw_output)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from manilaclient import api_versions
from osc_lib import exceptions
from osc_lib import utils as oscutils

from openstackclient.share.v2 import resource_locks
from openstackclient.tests.unit.share.v2 import fakes as share_fakes
from openstackclient.tests.unit import utils as test_utils

DETAIL_COLUMNS = [
    'ID',
    'Resource Id',
    'Resource Type',
    'Resource Action',
    'Created At',
    'Updated At',
    'User Id',
    'Project Id',
    'Lock Reason',
    'Lock Context',
]

SUMMARY_COLUMNS = [
    'ID',
    'Resource Id',
    'Resource Type',
    'Resource Action',
]


class TestResourceLock(share_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.shares_mock = self.share_client.shares
        self.shares_mock.reset_mock()

        self.locks_mock = self.share_client.resource_locks
        self.locks_mock.reset_mock()

        self.set_share_api_version(api_versions.MAX_VERSION)


class TestResourceLockCreate(TestResourceLock):
    def setUp(self):
        super().setUp()

        self.share = share_fakes.FakeShare.create_one_share()
        self.shares_mock.create.return_value = self.share

        self.shares_mock.get.return_value = self.share

        self.lock = share_fakes.FakeResourceLock.create_one_lock(
            attrs={'resource_id': self.share.id}
        )
        self.locks_mock.get.return_value = self.lock
        self.locks_mock.create.return_value = self.lock

        self.cmd = resource_locks.CreateResourceLock(self.app, None)

        self.data = tuple(self.lock._info.values())
        self.columns = tuple(self.lock._info.keys())

    def test_share_lock_create_missing_required_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_lock_create(self):
        arglist = [
            '--resource-action',
            'revert_to_snapshot',
            '--lock-reason',
            "you cannot go back in time",
            self.share.id,
            'share',
        ]
        verifylist = [
            ('resource', self.share.id),
            ('resource_type', 'share'),
            ('resource_action', 'revert_to_snapshot'),
            ('lock_reason', 'you cannot go back in time'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.locks_mock.create.assert_called_with(
            self.share.id,
            'share',
            'revert_to_snapshot',
            'you cannot go back in time',
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestResourceLockDelete(TestResourceLock):
    def setUp(self):
        super().setUp()

        self.lock = share_fakes.FakeResourceLock.create_one_lock()

        self.locks_mock.get.return_value = self.lock
        self.lock.delete = mock.Mock()

        self.cmd = resource_locks.DeleteResourceLock(self.app, None)

    def test_share_lock_delete_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_lock_delete(self):
        arglist = [self.lock.id]
        verifylist = [('lock', [self.lock.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.lock.delete.assert_called_once_with()
        self.assertIsNone(result)

    def test_share_lock_delete_multiple(self):
        locks = share_fakes.FakeResourceLock.create_locks(count=2)
        arglist = [locks[0].id, locks[1].id]
        verifylist = [('lock', [locks[0].id, locks[1].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertEqual(self.lock.delete.call_count, len(locks))
        self.assertIsNone(result)

    def test_share_lock_delete_exception(self):
        arglist = [self.lock.id]
        verifylist = [('lock', [self.lock.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.lock.delete.side_effect = exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestResourceLockShow(TestResourceLock):
    def setUp(self):
        super().setUp()

        self.lock = share_fakes.FakeResourceLock.create_one_lock()
        self.locks_mock.get.return_value = self.lock

        self.cmd = resource_locks.ShowResourceLock(self.app, None)

        self.data = self.lock._info.values()
        self.columns = list(self.lock._info.keys())

    def test_share_lock_show_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_lock_show(self):
        arglist = [
            self.lock.id,
        ]
        verifylist = [('lock', self.lock.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.locks_mock.get.assert_called_with(self.lock.id)
        self.assertEqual(len(self.columns), len(columns))
        self.assertCountEqual(sorted(self.data), sorted(data))


class TestResourceLockList(TestResourceLock):
    def setUp(self):
        super().setUp()

        self.locks = share_fakes.FakeResourceLock.create_locks(count=2)

        self.locks_mock.list.return_value = self.locks

        self.values = (
            oscutils.get_dict_properties(m._info, DETAIL_COLUMNS)
            for m in self.locks
        )

        self.cmd = resource_locks.ListResourceLock(self.app, None)

    def test_share_lock_list(self):
        arglist = ['--detailed']
        verifylist = [('detailed', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.locks_mock.list.assert_called_with(
            search_opts={
                'all_projects': False,
                'project_id': None,
                'user_id': None,
                'id': None,
                'resource_id': None,
                'resource_type': None,
                'resource_action': None,
                'lock_context': None,
                'created_before': None,
                'created_since': None,
                'limit': None,
                'offset': None,
            },
            sort_key=None,
            sort_dir=None,
        )

        self.assertEqual(sorted(DETAIL_COLUMNS), sorted(columns))
        actual_data = [sorted(d) for d in data]
        expected_data = [sorted(v) for v in self.values]
        self.assertEqual(actual_data, expected_data)


class TestResourceLockSet(TestResourceLock):
    def setUp(self):
        super().setUp()

        self.lock = share_fakes.FakeResourceLock.create_one_lock()
        self.lock.update = mock.Mock()

        self.locks_mock.get.return_value = self.lock

        self.cmd = resource_locks.SetResourceLock(self.app, None)

    def test_share_lock_set_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_lock_set(self):
        arglist = [
            self.lock.id,
            '--resource-action',
            'unmanage',
        ]
        verifylist = [('lock', self.lock.id), ('resource_action', 'unmanage')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.locks_mock.update.assert_called_with(
            self.lock.id, resource_action='unmanage'
        )


class TestResourceLockUnSet(TestResourceLock):
    def setUp(self):
        super().setUp()

        self.lock = share_fakes.FakeResourceLock.create_one_lock()
        self.lock.update = mock.Mock()

        self.locks_mock.get.return_value = self.lock

        self.cmd = resource_locks.UnsetResourceLock(self.app, None)

    def test_share_lock_unset_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_lock_unset(self):
        arglist = [self.lock.id, '--lock-reason']
        verifylist = [('lock', self.lock.id), ('lock_reason', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.locks_mock.update.assert_called_with(
            self.lock.id, lock_reason=None
        )
